"""
Tag reads by haplotype

Sequencing reads are read from file ALIGNMENTS (in BAM or CRAM format) and tagged reads
are written to stdout.
"""
import logging
import os
import sys
import pysam
import hashlib
from collections import defaultdict
from typing import List, Optional, Union, Dict, Tuple, FrozenSet, Sequence

from xopen import xopen

from contextlib import ExitStack
from whatshap import __version__
from whatshap.cli import PhasedInputReader, CommandLineError
from whatshap.vcf import VcfReader, VcfError, VariantTable, VariantCallPhase, VcfInvalidChromosome
from whatshap.core import NumericSampleIds
from whatshap.timer import StageTimer
from whatshap.utils import Region, stdout_is_regular_file


logger = logging.getLogger(__name__)


# fmt: off
def add_arguments(parser):
    arg = parser.add_argument
    arg("-o", "--output",
        default=sys.stdout,
        help="Output file. If omitted, use standard output.")
    arg("--reference", "-r", metavar="FASTA",
        help="Reference file. Must be accompanied by .fai index (create with samtools faidx)")
    arg("--no-reference", action="store_true", default=False,
        help="Detect alleles without requiring a reference, at the expense of phasing quality "
        "(in particular for long reads)")
    arg("--regions", dest="regions", metavar="REGION", default=None, action="append",
        help="Specify region(s) of interest to limit the tagging to reads/variants "
        "overlapping those regions. You can specify a space-separated list of "
        "regions in the form of chrom:start-end, chrom (consider entire chromosome), "
        "or chrom:start (consider region from this start to end of chromosome).")
    arg("--ignore-linked-read", default=False, action="store_true",
        help="Ignore linkage information stored in BX tags of the reads.")
    arg("--linked-read-distance-cutoff", "-d", metavar="LINKEDREADDISTANCE",
        default=50000, type=int,
        help="Assume reads with identical BX tags belong to different read clouds if their "
        "distance is larger than LINKEDREADDISTANCE (default: %(default)s).")
    arg("--ignore-read-groups", default=False, action="store_true",
        help="Ignore read groups in BAM/CRAM header and assume all reads come "
        "from the same sample.")
    arg("--sample", dest="given_samples", metavar="SAMPLE", default=None, action="append",
        help="Name of a sample to phase. If not given, all samples in the "
        "input VCF are phased. Can be used multiple times.")
    arg("--output-haplotag-list", dest="haplotag_list", metavar="HAPLOTAG_LIST", default=None,
        help="Write assignments of read names to haplotypes (tab separated) to given "
        "output file. If filename ends in .gz, then output is gzipped.")
    arg("--tag-supplementary", default=False, action="store_true",
        help="Also tag supplementary alignments. Supplementary alignments are assigned to the "
        "same haplotype as the primary alignment (default: only tag primary alignments).")
    arg("--skip-missing-contigs", default=False, action="store_true",
        help="Skip reads that map to a contig that does not exist in the VCF")
    arg("--output-threads", "--out-threads", default=1, type=int,
        help="Number of threads to use for output file writing (passed to pysam). "
        "For optimal performance, instead pipe output into 'samtools view' to compress.")
    arg("variant_file", metavar="VCF", help="VCF file with phased variants "
        "(must be gzip-compressed and indexed)")
    arg("alignment_file", metavar="ALIGNMENTS",
        help="BAM/CRAM file with alignments to be tagged by haplotype")
# fmt: on


def validate(args, parser):
    if args.reference is not None and args.no_reference:
        parser.error("Options --reference and --no-reference cannot be used together")


def md5_of(filename):
    with open(filename, "rb") as f:
        return hashlib.md5(f.read()).hexdigest()


def get_variant_information(variant_table: VariantTable, sample: str):
    """
    Return (vpos_to_phase_info, variants) pair, where vpos_to_phase_info maps
    a variant position to a pair (integer block_id, phase0),
    and variants is a list of all non-homozygous variants.
    """
    genotypes = variant_table.genotypes_of(sample)
    phases: List[Optional[VariantCallPhase]] = variant_table.phases_of(sample)

    vpos_to_phase_info = dict()
    variants = []
    for v, gt, phase in zip(variant_table.variants, genotypes, phases):
        if phase is None or phase.block_id is None:
            continue
        # assuming ploidy = 2
        phase_info = int(phase.block_id), phase.phase[0]
        vpos_to_phase_info[v.position] = phase_info
        if not gt.is_homozygous():
            variants.append(v)

    return vpos_to_phase_info, variants


def attempt_add_phase_information(
    alignment, read_to_haplotype, bxtag_to_haplotype, linked_read_cutoff, ignore_linked_read
):
    is_tagged = 0
    haplotype_name = "none"
    phaseset = "none"
    try:
        haplotype, quality, phaseset = read_to_haplotype[alignment.query_name]
        haplotype_name = f"H{haplotype + 1}"
        alignment.set_tag("HP", haplotype + 1)
        alignment.set_tag("PC", quality)
        alignment.set_tag("PS", phaseset)
        is_tagged = 1
    except KeyError:
        # check if reads with same tag have been assigned
        if not ignore_linked_read:
            try:
                tag = alignment.get_tag("BX")
            except KeyError:
                read_clouds = []
            else:  # alignment has BX tag
                read_clouds = bxtag_to_haplotype[tag]

            for (reference_start, haplotype, phaseset) in read_clouds:
                if abs(reference_start - alignment.reference_start) <= linked_read_cutoff:
                    haplotype_name = f"H{haplotype + 1}"
                    alignment.set_tag("HP", haplotype + 1)
                    alignment.set_tag("PC", value=None)
                    alignment.set_tag("PS", phaseset)
                    is_tagged = 1
                    break
    return is_tagged, haplotype_name, phaseset


def load_chromosome_variants(
    vcf_reader: VcfReader, chromosome: str, regions: Sequence[Tuple[int, Optional[int]]]
) -> VariantTable:
    try:
        logger.debug(f"Loading variants from {len(regions)} distinct region(s)")
        variant_table = vcf_reader.fetch_regions(chromosome, regions)
        logger.debug(f"Loaded {len(variant_table)} variants for chromosome {chromosome} in VCF")
    except OSError as err:
        # not entirely clear to me why this could raise
        # an OSError at this point?
        logger.error(str(err))
        raise err
    return variant_table


def prepare_haplotag_information(
    variant_table,
    shared_samples,
    phased_input_reader,
    regions,
    ignore_linked_read,
    linked_read_cutoff,
):
    """
    Read all reads for this chromosome once to create one core.ReadSet per sample.
    This allows to assign phase to paired-end reads based on both reads
    """
    n_multiple_phase_sets = 0
    BX_tag_to_haplotype = defaultdict(list)
    # maps read name to (haplotype, quality, phaseset)
    read_to_haplotype = {}

    for sample in shared_samples:
        variantpos_to_phaseinfo, variants = get_variant_information(variant_table, sample)
        read_set, _ = phased_input_reader.read(
            variant_table.chromosome, variants, sample, regions=regions
        )

        # map BX tag to list of reads
        bx_tag_to_readlist = defaultdict(list)
        if not ignore_linked_read:
            for read in read_set:
                if read.has_BX_tag():
                    bx_tag_to_readlist[read.BX_tag].append(read)

        # all reads processed so far
        processed_reads = set()
        for read in read_set:
            if read.name in processed_reads:
                continue
            # mapping: phaseset --> phred scaled difference between costs of assigning reads to haplotype 0 or 1
            haplotype_costs = defaultdict(int)

            processed_reads.add(read.name)
            reads_to_consider = {read}

            # reads with same BX tag need to be considered too (unless --ignore-linked-read is set)
            if not ignore_linked_read and read.has_BX_tag():
                for r in bx_tag_to_readlist[read.BX_tag]:
                    if r.name not in processed_reads:
                        # only select reads close to current one
                        if abs(read.reference_start - r.reference_start) <= linked_read_cutoff:
                            reads_to_consider.add(r)

            for r in reads_to_consider:
                processed_reads.add(r.name)
                for v in r:
                    assert v.allele in [0, 1]
                    phaseset, allele = variantpos_to_phaseinfo[v.position]
                    if v.allele == allele:
                        haplotype_costs[phaseset] += v.quality
                    else:
                        haplotype_costs[phaseset] -= v.quality

            l = list(haplotype_costs.items())
            l.sort(key=lambda t: -abs(t[1]))
            # logger.info('Read %s: %s', read.name, str(l))

            if len(l) == 0:
                continue
            if len(l) > 1:
                n_multiple_phase_sets += 1
            phaseset, quality = l[0]
            if quality == 0:
                continue
            haplotype = 0 if quality > 0 else 1

            if not ignore_linked_read and read.has_BX_tag():
                BX_tag_to_haplotype[read.BX_tag].append((read.reference_start, haplotype, phaseset))

            for r in reads_to_consider:
                read_to_haplotype[r.name] = (haplotype, abs(quality), phaseset)
                logger.debug(
                    "Assigned read {} to haplotype {} with a "
                    "quality of {} based on {} covered variants".format(
                        r.name, haplotype, quality, len(r)
                    )
                )
    return BX_tag_to_haplotype, read_to_haplotype, n_multiple_phase_sets


def normalize_user_regions(
    user_regions: Optional[Sequence[str]], bam_references: List[str]
) -> Dict[str, List[Tuple[int, Optional[int]]]]:
    """
    Process and accept user input of the following forms:

    chr:start-end -> chr, start, end
    chr -> chr, 0, None
    chr:start -> chr, start, None

    User input is interpreted as 1-based closed intervals.
    In pysam, coordinates are treated as 0-based half-open,
    so convert user input chr1:1-100 into chr1:0-100

    Args:
        user_regions: list of user input regions
        bam_references: references of BAM file

    Returns:
        dict of lists containing normalized regions per chromosome
    """
    regions: Dict[str, List[Tuple[int, Optional[int]]]] = defaultdict(list)
    if user_regions is None:
        for reference in bam_references:
            regions[reference].append((0, None))
    else:
        bam_references = set(bam_references)
        for region_spec in user_regions:
            region = Region.parse(region_spec)
            if region.chromosome not in bam_references:
                raise ValueError(
                    "Requested reference '{region.chromosome}' not found in input BAM/CRAM"
                )
            regions[region.chromosome].append((region.start, region.end))
    return regions


def compute_variant_file_samples_to_use(vcf_samples, user_given_samples, ignore_read_groups):
    """
    Open variant file and load sample information - check if samples in VCF are compatible
    with user specified list of samples.

    return iterable of VCF samples to use
    """
    samples_in_vcf = set(vcf_samples)
    if len(samples_in_vcf) < 1:
        raise VcfError("No samples detected in VCF file; cannot perform haplotagging")
    logger.info(f"Found {len(samples_in_vcf)} sample(s) in input VCF")
    logger.debug(
        "Found the following samples in input VCF: {}".format(" - ".join(sorted(samples_in_vcf)))
    )

    if ignore_read_groups and user_given_samples is None and len(samples_in_vcf) > 1:
        raise ValueError(
            'When setting "--ignore-read-groups" on '
            "a multi-sample VCF, samples to be used must "
            'be specified via the "--sample" parameter.'
        )

    if user_given_samples is None:
        samples_to_use = samples_in_vcf
    else:
        given_samples = user_given_samples
        missing_samples = set(given_samples) - samples_in_vcf
        if len(missing_samples) > 0:
            raise VcfError(
                "The following samples were specified via the "
                '"--sample" parameter, but are not part of the '
                "input VCF: {}".format(sorted(missing_samples))
            )

        samples_to_use = samples_in_vcf.intersection(given_samples)
        logger.info(f"Keeping {len(samples_to_use)} sample(s) for haplo-tagging")
        logger.debug(
            "Keeping the following samples for haplo-tagging: {}".format(
                " - ".join(sorted(samples_to_use))
            )
        )
    return samples_to_use


def compute_shared_samples(bam_reader, ignore_read_groups, vcf_samples):
    """
    Return final samples to use for haplo-tagging
    """
    read_groups = bam_reader.header.get("RG", [])
    bam_samples = {(rg["SM"] if "SM" in rg else "") for rg in read_groups}

    logger.info(f"Found {len(bam_samples)} sample(s) in BAM file")
    logger.debug(
        "Found the following samples in BAM file: {}".format(",".join(sorted(bam_samples)))
    )

    if not ignore_read_groups:
        shared_samples = bam_samples.intersection(vcf_samples)
        if len(shared_samples) == 0:
            raise ValueError(
                "No common samples between VCF and BAM file detected. "
                'You may restart the analysis setting "--ignore-read-groups" '
                "(if appropriate) to avoid this error."
            )
        elif len(shared_samples) < len(bam_samples):
            missing_samples = " | ".join(sorted(bam_samples - shared_samples))
            logger.warning(
                "Ignoring the following sample(s) for haplo-tagging "
                "because they are not part of the VCF or "
                'were not requested via "--sample": {}'.format(missing_samples)
            )
        else:
            # situation is ok
            pass
    else:
        shared_samples = vcf_samples
    return shared_samples


def open_output_alignment_file(aln_output, reference, vcf_md5, bam_header, threads=1):
    """
    :param aln_output:
    :param reference:
    :param vcf_md5:
    :param bam_header:
    :param exit_stack:
    :return:
    """
    # Prepare header
    # TODO: convince pysam to allow @HS header line
    command_line = " ".join(["whatshap"] + sys.argv[1:])
    PG_entry = {
        "ID": "whatshap",
        "PN": "whatshap",
        "VN": __version__,
        "CL": command_line,
        "m5": vcf_md5,
    }
    if "PG" in bam_header:
        bam_header["PG"].append(PG_entry)
    else:
        bam_header["PG"] = [PG_entry]
    if aln_output is None:
        aln_output = "-"
        kwargs = dict()
    elif str(aln_output).endswith(".cram"):  # FIXME hard-coded value
        if reference is None:
            raise ValueError(
                'Writing CRAM output requires FASTA reference file given via "--reference"'
            )
        kwargs = dict(mode="wc", reference_filename=reference)
    else:
        # Write BAM, disable compression when piping
        if aln_output is sys.stdout and not stdout_is_regular_file():
            kwargs = dict(mode="wb0", threads=threads)
        else:
            kwargs = dict(mode="wb", threads=threads)
    try:
        bam_writer = pysam.AlignmentFile(
            aln_output, header=pysam.AlignmentHeader.from_dict(bam_header), **kwargs
        )
    except OSError as err:
        raise CommandLineError(
            f"Error while initializing alignment output file at path: {aln_output}\n{err}"
        )

    return bam_writer


def open_haplotag_writer(path):
    if path is None:
        path = os.devnull
    try:
        writer = xopen(path, "wt")
    except OSError as err:
        raise CommandLineError(
            f"Error while initializing haplotag list output at path: {path}\n{err}"
        )
    logger.debug("Writing header line to haplotag list output file")
    print("#readname", "haplotype", "phaseset", "chromosome", sep="\t", file=writer)
    return writer


def ignore_read(alignment, tag_supplementary):
    """
    If supplementary alignments should also be tagged,
    this should only take the haplo-tag of the primary
    alignment into account - this leads to:

    We ignore an alignment [aln]:
    - IF aln is_unmapped OR is_secondary
    - IF tag_supplementary AND aln is_secondary
    - IF not tag_supplementary AND is_supplementary

    :param alignment:
    :param tag_supplementary:
    :return:
    """
    # TODO: could be that some checks here are not needed
    # due to default filtering in ReadSetReader::_usableAlignments

    if alignment.is_unmapped or alignment.is_secondary:
        # unmapped or secondary alignments are never tagged
        ignore = True
    elif tag_supplementary and alignment.is_supplementary:
        # from the previous if, we know
        # the alignment to be primary
        ignore = False
    elif alignment.is_supplementary:
        # tag_supplementary is False, so discard
        ignore = True
    else:
        # whatever is left should be good
        ignore = False
    return ignore


def contigs_with_alignments(af: pysam.AlignmentFile) -> FrozenSet[str]:
    has_alignments = []
    for contig in af.references:
        for _ in af.fetch(contig=contig):
            has_alignments.append(contig)
            break
    return frozenset(has_alignments)


def run_haplotag(
    variant_file,
    alignment_file,
    output=None,
    reference: Union[None, bool, str] = False,
    regions=None,
    ignore_linked_read=False,
    given_samples=None,
    linked_read_distance_cutoff=50000,
    ignore_read_groups=False,
    haplotag_list=None,
    tag_supplementary=False,
    skip_missing_contigs=False,
    output_threads=1,
):

    timers = StageTimer()
    timers.start("haplotag-run")

    if output in (None, sys.stdout) and sys.stdout.isatty():
        raise CommandLineError(
            "Refusing to write BAM to the terminal. Either use the '-o' option or redirect "
            "standard output with '>'."
        )
    with ExitStack() as stack:
        timers.start("haplotag-init")
        try:
            vcf_reader = stack.enter_context(VcfReader(variant_file, indels=True, phases=True))
        except OSError as err:
            raise CommandLineError(f"Error while loading variant file {variant_file}: {err}")

        use_vcf_samples = compute_variant_file_samples_to_use(
            vcf_reader.samples, given_samples, ignore_read_groups
        )

        try:
            bam_reader = stack.enter_context(
                pysam.AlignmentFile(alignment_file, "rb", require_index=True)
            )
        except OSError as err:
            raise CommandLineError(f"Error while loading alignment file {alignment_file}: {err}")
        # This checks also sample compatibility with VCF
        shared_samples = compute_shared_samples(bam_reader, ignore_read_groups, use_vcf_samples)

        # Check if user has specified a subset of regions per chromosome
        user_regions = normalize_user_regions(regions, bam_reader.references)

        include_unmapped = regions is None
        phased_input_reader = stack.enter_context(
            PhasedInputReader(
                [alignment_file],
                None if reference is False else reference,
                NumericSampleIds(),
                ignore_read_groups,
                indels=False,
            )
        )
        if phased_input_reader.has_alignments and reference is None:
            raise CommandLineError(
                "A reference FASTA needs to be provided with -r/--reference; "
                "or use --no-reference at the expense of phasing quality."
            )

        bam_writer = stack.enter_context(
            open_output_alignment_file(
                output,
                reference,
                md5_of(variant_file),
                bam_reader.header.to_dict(),
                threads=output_threads,
            )
        )
        haplotag_writer = stack.enter_context(open_haplotag_writer(haplotag_list))

        timers.stop("haplotag-init")
        logger.debug(
            "All input/output files initialized (time: {})".format(timers.elapsed("haplotag-init"))
        )
        timers.start("haplotag-process")

        n_alignments = 0
        n_tagged = 0
        n_multiple_phase_sets = 0

        has_alignments = contigs_with_alignments(bam_reader)

        for chrom, regions in user_regions.items():
            logger.debug(f"Processing chromosome {chrom}")

            if chrom not in has_alignments:
                # Skip chromosomes without alignments. This allows to have extra chromosomes in the
                # BAM header compared to the VCF.
                continue
            try:
                variant_table = load_chromosome_variants(vcf_reader, chrom, regions)
            except VcfInvalidChromosome:
                if skip_missing_contigs:
                    logger.info(
                        f"Skipping reads on '{chrom}' because the contig does not exist in the VCF"
                    )
                    continue
                else:
                    raise CommandLineError(
                        f"Input BAM/CRAM contains reads on contig '{chrom}', but that contig does "
                        "not exist in the VCF header. To bypass this check, use "
                        "--skip-missing-contigs"
                    )
            except VcfError as e:
                raise CommandLineError(str(e))
            if variant_table is not None:
                logger.debug("Preparing haplotype information")
                (BX_tag_to_haplotype, read_to_haplotype, n_mult) = prepare_haplotag_information(
                    variant_table,
                    shared_samples,
                    phased_input_reader,
                    regions,
                    ignore_linked_read,
                    linked_read_distance_cutoff,
                )
                n_multiple_phase_sets += n_mult
            else:
                # avoid uninitialized variables
                BX_tag_to_haplotype = None
                read_to_haplotype = None

            assert not include_unmapped or len(regions) == 1
            for start, end in regions:
                logger.debug("Working on %s:%d-%d", chrom, start, end)
                for alignment in bam_reader.fetch(contig=chrom, start=start, stop=end):
                    n_alignments += 1
                    haplotype_name = "none"
                    phaseset = "none"

                    if variant_table is None or ignore_read(alignment, tag_supplementary):
                        # - If no variants in VCF for this chromosome,
                        # alignments just get written to output
                        # - Ignored reads are simply
                        # written to the output BAM
                        # Existing tags HP, PC and PS are removed
                        alignment.set_tag("HP", value=None)
                        alignment.set_tag("PC", value=None)
                        alignment.set_tag("PS", value=None)
                    else:
                        (is_tagged, haplotype_name, phaseset) = attempt_add_phase_information(
                            alignment,
                            read_to_haplotype,
                            BX_tag_to_haplotype,
                            linked_read_distance_cutoff,
                            ignore_linked_read,
                        )
                        n_tagged += is_tagged

                        if not is_tagged:
                            # Remove any existing tags HP, PC and PS if the aligment does
                            # not have phasing information
                            alignment.set_tag("HP", value=None)
                            alignment.set_tag("PC", value=None)
                            alignment.set_tag("PS", value=None)

                    bam_writer.write(alignment)
                    if haplotag_list is not None and not (
                        alignment.is_secondary or alignment.is_supplementary
                    ):
                        print(
                            alignment.query_name,
                            haplotype_name,
                            phaseset,
                            chrom,
                            sep="\t",
                            file=haplotag_writer,
                        )

                    if n_alignments % 100000 == 0:
                        logger.debug(f"Processed {n_alignments} alignment records.")
        if include_unmapped:
            for alignment in bam_reader.fetch(until_eof=True):
                bam_writer.write(alignment)
        timers.stop("haplotag-process")
        logger.debug("Processing complete (time: {})".format(timers.elapsed("haplotag-process")))

    timers.stop("haplotag-run")

    logger.info("\n== SUMMARY ==")
    logger.info("Total alignments processed:              %12d", n_alignments)
    logger.info("Alignments that could be tagged:         %12d", n_tagged)
    logger.info("Alignments spanning multiple phase sets: %12d", n_multiple_phase_sets)
    logger.info("Finished in %.1f s", timers.elapsed("haplotag-run"))


def main(args):
    if args.no_reference:
        args.reference = False
    del args.no_reference
    run_haplotag(**vars(args))



from argparse import ArgumentParser, RawDescriptionHelpFormatter
import sys


class HelpfulArgumentParser(ArgumentParser):
    """An ArgumentParser that prints full help on errors."""

    def __init__(self, *args, **kwargs):
        if "formatter_class" not in kwargs:
            kwargs["formatter_class"] = RawDescriptionHelpFormatter
        super().__init__(*args, **kwargs)

    def error(self, message):
        self.print_help(sys.stderr)
        args = {"prog": self.prog, "message": message}
        self.exit(2, "%(prog)s: error: %(message)s\n" % args)



"""
Test Threading
"""

from whatshap.core import Read, ReadSet

# from collections import defaultdict
from whatshap.threading import get_allele_depths, select_clusters
from whatshap.polyphaseutil import get_position_map, get_coverage

# from whatshap.reorder import compute_cut_positions


def create_testinstance1():
    var_pos = [
        24,
        56,
        89,
        113,
        162,
        166,
        187,
        205,
        211,
        248,
        273,
        299,
        307,
        324,
        351,
        370,
        378,
        400,
        441,
        455,
        478,
        492,
    ]
    readset = ReadSet()
    matrix = [
        "0011000",
        "11010100",
        " 101011010",
        " 0001011000",
        "  11001001",
        "  0010100000",
        "   100010001",
        "       0100000101",
        "    101110001",
        "        0001110011",
        "        1010001010",
        "     011100011",
        "         0010100111",
        "          1010101011",
        "          0101001110",
        "              01000001",
        "              01010001",
        "                101100",
        "                111010",
    ]
    for i in range(len(matrix)):
        read = Read(name="read" + str(i), mapq=15)
        for j in range(len(matrix[i])):
            if matrix[i][j] != " ":
                read.add_variant(var_pos[j], int(matrix[i][j]), 0)
        readset.add(read)
    clustering = [
        [0, 4, 6],
        [1, 2],
        [7, 10, 13],
        [9, 12, 14],
        [3, 5, 8, 11],
        [15, 16],
        [17],
        [18],
    ]
    genotypes = [
        {0: 2, 1: 1},
        {0: 2, 1: 1},
        {0: 2, 1: 1},
        {0: 1, 1: 2},
        {0: 2, 1: 1},
        {0: 2, 1: 1},
        {0: 2, 1: 1},
        {0: 2, 1: 1},
        {0: 2, 1: 1},
        {0: 3, 1: 0},
        {0: 2, 1: 1},
        {0: 2, 1: 1},
        {0: 2, 1: 1},
        {0: 1, 1: 2},
        {0: 2, 1: 1},
        {0: 2, 1: 1},
        {0: 1, 1: 2},
        {0: 2, 1: 1},
        {0: 1, 1: 2},
        {0: 2, 1: 1},
        {0: 2, 1: 1},
        {0: 2, 1: 1},
    ]
    return readset, var_pos, clustering, genotypes


def create_testinstance2():
    var_pos = [0, 1, 2, 3, 5, 8, 13, 21, 34, 55]
    readset = ReadSet()
    matrix = [
        "0010020000",
        "0000021000",
        "0010021000",
        "1020100010",
        "1020200010",
        "1020100010",
        "0101000002",
        "0101000112",
        "0101000002",
        "2000030100",
        "2001030100",
        "2000030100",
    ]
    for i in range(len(matrix)):
        read = Read(name="read" + str(i), mapq=15)
        for j in range(len(matrix[i])):
            if matrix[i][j] != " ":
                read.add_variant(var_pos[j], int(matrix[i][j]), 0)
        readset.add(read)
    clustering = [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9, 10, 11]]
    genotypes = [
        {0: 2, 1: 1, 2: 2},
        {0: 3, 1: 1},
        {0: 2, 1: 1, 2: 2},
        {0: 3, 1: 1},
        {0: 3, 1: 1},
        {0: 2, 2: 1, 3: 1},
        {0: 3, 1: 1},
        {0: 3, 1: 1},
        {0: 3, 2: 1},
        {0: 3, 1: 1},
    ]
    return readset, var_pos, clustering, genotypes


def create_testinstance3():
    var_pos = [0, 1, 2, 3, 5, 8, 13, 21, 34, 55]
    readset = ReadSet()
    matrix = [
        "011010 110",
        "011010 110",
        "011010 110",
        "011010 110",
        "011010 110",
        "011010 110",
        "011010 110",
        "011010 110",
        "011010 110",
        "0  1010001",
        "0  1010001",
        "0  1010001",
        "0  1010001",
        "0  1010001",
        "0  1010001",
        "0  1010001",
        "0  1010001",
        "1000000000",
        "1  0000",
        "0000000000",
    ]
    for i in range(len(matrix)):
        read = Read(name="read" + str(i), mapq=15)
        for j in range(len(matrix[i])):
            if matrix[i][j] != " ":
                read.add_variant(var_pos[j], int(matrix[i][j]), 0)
        readset.add(read)
    clustering = [list(range(9)), list(range(9, 17)), [17, 18], [19]]
    genotypes = [
        {0: 2},
        {1: 1},
        {1: 1},
        {0: 1, 1: 1},
        {0: 1, 1: 1},
        {0: 1, 1: 1},
        {0: 1},
        {0: 1, 1: 1},
        {0: 1, 1: 1},
        {0: 1, 1: 1},
    ]
    return readset, var_pos, clustering, genotypes


def test_position_map():
    for f in [create_testinstance1, create_testinstance2, create_testinstance3]:
        readset, var_pos, _, _ = f()
        index, rev_index = get_position_map(readset)
        for i, pos in enumerate(var_pos):
            assert index[pos] == i
            assert rev_index[i] == pos


def test_relative_coverage():
    readset, var_pos, clustering, _ = create_testinstance1()
    cov = get_coverage(readset, clustering)
    assert cov[0] == {0: 0.5, 1: 0.5}
    assert cov[1] == {0: 0.25, 1: 0.5, 4: 0.25}
    assert cov[2] == {0: 1 / 3, 1: 1 / 3, 4: 1 / 3}
    assert cov[3] == {0: 3 / 7, 1: 2 / 7, 4: 2 / 7}
    assert cov[4] == {0: 3 / 8, 1: 2 / 8, 4: 3 / 8}
    assert cov[5] == {0: 3 / 9, 1: 2 / 9, 4: 4 / 9}
    assert cov[6] == {0: 3 / 9, 1: 2 / 9, 4: 4 / 9}
    assert cov[7] == {0: 2 / 9, 1: 2 / 9, 2: 1 / 9, 4: 4 / 9}
    assert cov[8] == {0: 2 / 10, 1: 1 / 10, 2: 2 / 10, 3: 1 / 10, 4: 4 / 10}
    assert cov[9] == {0: 2 / 11, 1: 1 / 11, 2: 2 / 11, 3: 2 / 11, 4: 4 / 11}
    assert cov[10] == {0: 1 / 11, 2: 3 / 11, 3: 3 / 11, 4: 4 / 11}
    assert cov[11] == {0: 1 / 10, 2: 3 / 10, 3: 3 / 10, 4: 3 / 10}
    assert cov[12] == {2: 3 / 8, 3: 3 / 8, 4: 2 / 8}
    assert cov[13] == {2: 3 / 7, 3: 3 / 7, 4: 1 / 7}
    assert cov[14] == {2: 3 / 8, 3: 3 / 8, 5: 2 / 8}
    assert cov[15] == {2: 3 / 8, 3: 3 / 8, 5: 2 / 8}
    assert cov[16] == {2: 3 / 10, 3: 3 / 10, 5: 2 / 10, 6: 1 / 10, 7: 1 / 10}
    assert cov[17] == {2: 2 / 9, 3: 3 / 9, 5: 2 / 9, 6: 1 / 9, 7: 1 / 9}
    assert cov[18] == {2: 1 / 7, 3: 2 / 7, 5: 2 / 7, 6: 1 / 7, 7: 1 / 7}
    assert cov[19] == {2: 1 / 6, 3: 1 / 6, 5: 2 / 6, 6: 1 / 6, 7: 1 / 6}
    assert cov[20] == {5: 2 / 4, 6: 1 / 4, 7: 1 / 4}
    assert cov[21] == {5: 2 / 4, 6: 1 / 4, 7: 1 / 4}


def test_allele_depths():
    for f in [create_testinstance1, create_testinstance2, create_testinstance3]:
        readset, var_pos, clustering, genotypes = f()
        index, rev_index = get_position_map(readset)
        ploidy = sum(genotypes[0].values())
        ad, cons_lists = get_allele_depths(readset, clustering, ploidy=ploidy)
        for pos in range(len(rev_index)):
            for cid in range(len(clustering)):
                for al in [0, 1, 2, 3]:
                    val = sum(
                        [
                            1 if var.allele == al and var.position == rev_index[pos] else 0
                            for rid in clustering[cid]
                            for var in readset[rid]
                        ]
                    )
                    assert cid not in ad[pos] or al not in ad[pos][cid] or ad[pos][cid][al] == val


def test_cluster_selection1():
    readset, var_pos, clustering, genotypes = create_testinstance1()
    ad, cons_lists = get_allele_depths(readset, clustering, ploidy=3)
    c = select_clusters(ad, ploidy=3, max_gap=0)
    assert c[0] == [0, 1]
    assert c[1] == c[2] == c[3] == c[4] == c[5] == c[6] == [0, 1, 4]
    assert c[7] == [0, 1, 2, 4]
    assert c[8] == c[9] == [0, 1, 2, 3, 4]
    assert c[10] == c[11] == [0, 2, 3, 4]
    assert c[12] == c[13] == [2, 3, 4]
    assert c[14] == c[15] == [2, 3, 5]
    assert c[16] == c[17] == c[18] == c[19] == [2, 3, 5, 6, 7]
    assert c[20] == c[21] == [5, 6, 7]
    assert c == select_clusters(ad, ploidy=3, max_gap=1)


def test_cluster_selection2():
    readset, var_pos, clustering, genotypes = create_testinstance2()
    ad, cons_lists = get_allele_depths(readset, clustering, ploidy=4)
    c = select_clusters(ad, ploidy=4, max_gap=0)
    assert all([c[i] == [0, 1, 2, 3] for i in range(10)])
    assert c == select_clusters(ad, ploidy=3, max_gap=1)


def test_cluster_selection3():
    readset, var_pos, clustering, genotypes = create_testinstance3()
    ad, cons_lists = get_allele_depths(readset, clustering, ploidy=2)
    c = select_clusters(ad, ploidy=2, max_gap=0)
    assert c[0] == c[3] == c[4] == c[5] == [0, 1, 2]
    assert c[1] == c[2] == [0, 2, 3]
    assert c[6] == [1, 2, 3]
    assert c[7] == c[8] == c[9] == [0, 1]
    c = select_clusters(ad, ploidy=2, max_gap=1)
    assert c[0] == c[3] == c[4] == c[5] == [0, 1, 2]
    assert c[1] == c[2] == [0, 2, 3]
    assert c[6] == [0, 1, 2, 3]
    assert c[7] == c[8] == c[9] == [0, 1]
    c = select_clusters(ad, ploidy=2, max_gap=2)
    assert c[0] == c[3] == c[4] == c[5] == [0, 1, 2]
    assert c[1] == c[2] == c[6] == [0, 1, 2, 3]
    assert c[7] == c[8] == c[9] == [0, 1]
    c = select_clusters(ad, ploidy=2, max_gap=3)
    assert c[0] == [0, 1, 2]
    assert c[1] == c[2] == c[3] == c[4] == c[5] == c[6] == [0, 1, 2, 3]
    assert c[7] == c[8] == c[9] == [0, 1]

    assert c == select_clusters(ad, ploidy=2, max_gap=4)


"""
def test_cut_positions():
    path = [
        [2, 3, 5, 1],
        [2, 3, 5, 1],
        [2, 3, 6, 1],
        [2, 3, 6, 1],
        [4, 8, 6, 1],
        [5, 8, 6, 1],
        [5, 8, 6, 1],
        [5, 8, 6, 6],
        [8, 8, 6, 6],
        [8, 8, 6, 7],
        [8, 8, 6, 7],
        [8, 8, 9, 10],
        [8, 11, 9, 10],
        [8, 11, 9, 10],
    ]

    cuts1 = compute_cut_positions(path, 1, 12)
    cuts2 = compute_cut_positions(path, 2, 12)
    cuts3 = compute_cut_positions(path, 3, 12)
    cuts4 = compute_cut_positions(path, 4, 12)
    cuts5 = compute_cut_positions(path, 5, 12)

    assert cuts1[0] == [0]
    assert cuts2[0] == [0]
    assert cuts3[0] == [0, 4, 11]
    assert cuts4[0] == [0, 4, 9, 11]
    assert cuts5[0] == [0, 2, 4, 5, 7, 8, 9, 11, 12]

    assert cuts1[1] == [[0], [0], [0], [0]]
    assert cuts2[1] == [[0], [0], [0], [0]]
    assert cuts3[1] == [[0, 4], [0, 4], [0, 11], [0, 11]]
    assert cuts4[1] == [[0, 4], [0, 4], [0, 9, 11], [0, 9, 11]]
    assert cuts5[1] == [[0, 4, 5, 8, 12], [0, 4, 12], [0, 2, 9, 11], [0, 7, 9, 11]]


def test_multiswitch_improvement():
    path = [
        [3, 1, 2, 4],
        [3, 1, 2, 4],
        [5, 1, 2, 4],
        [5, 1, 2, 4],
        [5, 7, 6, 4],
        [5, 7, 6, 4],
        [5, 7, 6, 7],
        [5, 7, 6, 4],
        [5, 7, 6, 4],
        [8, 9, 10, 4],
        [8, 9, 10, 4],
    ]
    cluster_sim = [defaultdict(float) for _ in range(len(path))]
    for i in range(len(path)):
        cluster_sim[i][(1, 7)] = 0.6
        cluster_sim[i][(1, 6)] = 0.7
        cluster_sim[i][(2, 7)] = 0.8
        cluster_sim[i][(2, 6)] = 0.65
        cluster_sim[i][(5, 8)] = 0.3
        cluster_sim[i][(5, 9)] = 0.5
        cluster_sim[i][(5, 10)] = 0.8
        cluster_sim[i][(7, 8)] = 0.5
        cluster_sim[i][(7, 9)] = 0.8
        cluster_sim[i][(7, 10)] = 0.85
        cluster_sim[i][(6, 8)] = 0.9
        cluster_sim[i][(6, 9)] = 0.9
        cluster_sim[i][(6, 10)] = 0.6

    corrected_path = improve_path_on_multiswitches(path, 11, cluster_sim)

    truth = [
        [3, 1, 2, 4],
        [3, 1, 2, 4],
        [5, 1, 2, 4],
        [5, 1, 2, 4],
        [5, 6, 7, 4],
        [5, 6, 7, 4],
        [5, 6, 7, 7],
        [5, 6, 7, 4],
        [5, 6, 7, 4],
        [10, 8, 9, 4],
        [10, 8, 9, 4],
    ]

    for i in range(len(truth)):
        assert corrected_path[i] == truth[i]


def test_path_no_affine():
    readset, var_pos, clustering, genotypes = create_testinstance1()
    ploidy = 3

    index, rev_index = get_position_map(readset)
    num_vars = len(rev_index)
    positions = get_cluster_start_end_positions(readset, clustering, index)
    coverage = get_coverage(readset, clustering, index)
    cov_map = get_pos_to_clusters_map(coverage, ploidy)
    consensus = get_local_cluster_consensus(readset, clustering, cov_map, positions)
    allele_depths, cons = get_allele_depths(readset, clustering, cov_map)

    path = compute_threading_path(
        readset, num_vars, cov_map, allele_depths, ploidy, genotypes, affine_switch_cost=0.0,
    )
    cluster_paths = ["".join([str(path[i][j]) for i in range(len(path))]) for j in range(3)]

    first_block = set([cluster_paths[0][:20], cluster_paths[1][:20], cluster_paths[2][:20]])
    first_truth = set(["00000000003333333333", "11111111222222222222", "04444444444444555555"])
    second_block = set([cluster_paths[0][20:], cluster_paths[1][20:], cluster_paths[2][20:]])
    second_truth = set(["66", "77", "55"])

    print(cluster_paths)

    assert first_block == first_truth
    assert second_block == second_truth


def test_path_with_affine():
    readset, var_pos, clustering, genotypes = create_testinstance1()
    ploidy = 3

    index, rev_index = get_position_map(readset)
    num_vars = len(rev_index)
    positions = get_cluster_start_end_positions(readset, clustering, index)
    coverage = get_coverage(readset, clustering, index)
    cov_map = get_pos_to_clusters_map(coverage, ploidy)
    consensus = get_local_cluster_consensus(readset, clustering, cov_map, positions)
    allele_depths, cons = get_allele_depths(readset, clustering, cov_map)

    path = compute_threading_path(readset, num_vars, cov_map, allele_depths, ploidy, genotypes)
    cluster_paths = ["".join([str(path[i][j]) for i in range(len(path))]) for j in range(3)]

    first_block = set([cluster_paths[0][:9], cluster_paths[1][:9], cluster_paths[2][:9]])
    first_truth = set(["000000000", "111111111", "044444444"])
    second_block = set([cluster_paths[0][9:20], cluster_paths[1][9:20], cluster_paths[2][9:20]])
    second_truth = set(["33333333333", "22222222222", "44444555555"])
    third_block = set([cluster_paths[0][20:], cluster_paths[1][20:], cluster_paths[2][20:]])
    third_truth = set(["66", "77", "55"])

    print(cluster_paths)

    assert first_block == first_truth
    assert second_block == second_truth
    assert third_block == third_truth
    """



"""
Edit distance tests

The initial version was copied from https://bitbucket.org/marcelm/sqt/src/af255d54a21815cb9a3e0b279b431a320d4626bd/tests/testalign.py
"""
from whatshap.align import edit_distance as ed
from whatshap.align import edit_distance_affine_gap as ed_aff
from random import choice, seed, randint


STRING_PAIRS = [
    ("", ""),
    ("", "A"),
    ("A", "A"),
    ("AB", ""),
    ("AB", "ABC"),
    ("TGAATCCC", "CCTGAATC"),
    ("ANANAS", "BANANA"),
    ("SISSI", "MISSISSIPPI"),
    ("GGAATCCC", "TGAGGGATAAATATTTAGAATTTAGTAGTAGTGTT"),
    ("TCTGTTCCCTCCCTGTCTCA", "TTTTAGGAAATACGCC"),
    (
        "TGAGACACGCAACATGGGAAAGGCAAGGCACACAGGGGATAGG",
        "AATTTATTTTATTGTGATTTTTTGGAGGTTTGGAAGCCACTAAGCTATACTGAGACACGCAACAGGGGAAAGGCAAGGCACA",
    ),
    (
        "TCCATCTCATCCCTGCGTGTCCCATCTGTTCCCTCCCTGTCTCA",
        "TTTTAGGAAATACGCCTGGTGGGGTTTGGAGTATAGTGAAAGATAGGTGAGTTGGTCGGGTG",
    ),
    ("A", "TCTGCTCCTGGCCCATGATCGTATAACTTTCAAATTT"),
    ("GCGCGGACT", "TAAATCCTGG"),
]


def randstring():
    return "".join(choice("AC") for _ in range(randint(0, 10)))


seed(10)
STRING_PAIRS.extend((randstring(), randstring()) for _ in range(1000))


def test_edit_distance():
    assert ed("", "") == 0
    assert ed("", "A") == 1
    assert ed("A", "B") == 1
    assert ed("A", "A") == 0
    assert ed("A", "AB") == 1
    assert ed("BA", "AB") == 2
    for s, t in STRING_PAIRS:
        assert ed(s, "") == len(s)
        assert ed("", s) == len(s)
        assert ed(s, t) == ed(t, s)


def test_edit_distance_bytes():
    assert ed(b"", b"") == 0
    assert ed(b"", b"A") == 1
    assert ed(b"A", b"B") == 1
    assert ed(b"A", b"A") == 0
    assert ed(b"A", b"AB") == 1
    assert ed(b"BA", b"AB") == 2
    for s, t in STRING_PAIRS:
        s = s.encode("ascii")
        t = t.encode("ascii")
        assert ed(s, b"") == len(s)
        assert ed(b"", s) == len(s)
        assert ed(s, t) == ed(t, s)


def assert_banded(s, t, maxdiff):
    banded_dist = ed(s, t, maxdiff=maxdiff)
    true_dist = ed(s, t)
    if true_dist > maxdiff:
        assert banded_dist > maxdiff
    else:
        assert banded_dist == true_dist


def test_edit_distance_banded():
    for maxdiff in range(5):
        assert_banded("ABC", "", maxdiff)

        for s, t in STRING_PAIRS:
            assert_banded(s, "", maxdiff)
            assert_banded("", s, maxdiff)
            assert_banded(s, t, maxdiff)
            assert_banded(t, s, maxdiff)


def test_edit_distance_affine():
    for mismatch_cost in [1, 10, 30, 40, 50]:
        for gap_start in [1, 10, 30, 40, 50]:
            assert ed_aff("", "", [], gap_start, 10) == 0
            assert ed_aff("", "A", [], gap_start, 10) == gap_start
            assert ed_aff("A", "B", [mismatch_cost], gap_start, 10) == min(
                gap_start * 2, mismatch_cost
            )
            assert ed_aff("A", "A", [mismatch_cost], gap_start, 10) == 0
            assert ed_aff("A", "AB", [mismatch_cost], gap_start, 10) == gap_start
            assert ed_aff("BA", "AB", [mismatch_cost] * 2, gap_start, 100) == min(
                2 * mismatch_cost, 2 * gap_start
            )
            for s, t in STRING_PAIRS:
                if s != "":
                    assert (
                        ed_aff(s, "", [mismatch_cost] * len(s), gap_start, 10)
                        == gap_start + (len(s) - 1) * 10
                    )
                    assert ed_aff("", s, [], gap_start, 10) == gap_start + (len(s) - 1) * 10
                assert ed_aff(s, t, [mismatch_cost] * len(s), gap_start, 10) == ed_aff(
                    t, s, [mismatch_cost] * len(t), gap_start, 10
                )


def test_edit_distance_affine_bytes():
    for mismatch_cost in [1, 10, 20, 30, 40, 50]:
        for gap_start in [1, 10, 20, 30, 40, 50]:
            assert ed_aff(b"", b"", [], gap_start, 10) == 0
            assert ed_aff(b"", b"A", [], gap_start, 10) == gap_start
            assert ed_aff(b"A", b"B", [mismatch_cost], gap_start, 10) == min(
                gap_start * 2, mismatch_cost
            )
            assert ed_aff(b"A", b"A", [mismatch_cost], gap_start, 10) == 0
            assert ed_aff(b"A", b"AB", [mismatch_cost], gap_start, 10) == gap_start
            assert ed_aff(b"BA", b"AB", [mismatch_cost] * 2, gap_start, 100) == min(
                2 * mismatch_cost, 2 * gap_start
            )
            for s, t in STRING_PAIRS:
                s = s.encode("ascii")
                t = t.encode("ascii")
                if s != b"":
                    assert (
                        ed_aff(s, b"", [mismatch_cost] * len(s), gap_start, 10)
                        == gap_start + (len(s) - 1) * 10
                    )
                    assert ed_aff(b"", s, [], gap_start, 10) == gap_start + (len(s) - 1) * 10
                assert ed_aff(s, t, [mismatch_cost] * len(s), gap_start, 10) == ed_aff(
                    t, s, [mismatch_cost] * len(t), gap_start, 10
                )


def test_mismatches():
    for i in range(10):
        rand_costs = [randint(10, 70) for j in range(5)]
        assert ed_aff("AAAAA", "TTTTT", rand_costs, 100, 100) == sum(rand_costs)
        assert ed_aff("ATGCT", "ATCCT", rand_costs, 100, 100) == rand_costs[2]
        assert ed_aff("ATGGA", "ATGTTCA", rand_costs, 80, 10) == rand_costs[3] + 80 + 10


def test_small_examples():
    assert ed_aff("AGTCCGGTG", "AGTCCATCGGTC", [30, 40, 20, 20, 50, 60, 10, 20, 5], 40, 10) == 65
    assert ed_aff("ATGGCCG", "ATCGCTG", [40, 50, 10, 40, 50, 10, 40], 20, 10) == 20
    assert ed_aff("ATCCTC", "ATCGGGCTC", [50] * 6, 10, 5) == 20


def test_compare_to_edit_dist():
    for s, t in STRING_PAIRS:
        assert ed(s, t) == ed_aff(s, t, [1] * len(s), 1, 1)



from whatshap.cli.hapcut2vcf import run_hapcut2vcf


def test_hapcut2vcf(tmp_path):
    out = tmp_path / "hapcut.vcf"
    run_hapcut2vcf(
        hapcut="tests/data/pacbio/hapcut.txt", vcf="tests/data/pacbio/variants.vcf", output=out
    )



"""
Genotype variants

Runs only the genotyping algorithm. Genotype Likelihoods are computed using the
forward backward algorithm.
"""
import logging
import sys
import platform
from typing import Sequence

from contextlib import ExitStack
from whatshap import __version__
from whatshap.vcf import VcfReader, GenotypeVcfWriter
from whatshap.core import (
    ReadSet,
    Pedigree,
    NumericSampleIds,
    PhredGenotypeLikelihoods,
    GenotypeDPTable,
    compute_genotypes,
    Genotype,
)
from whatshap.pedigree import (
    PedReader,
    UniformRecombinationCostComputer,
    GeneticMapRecombinationCostComputer,
)
from whatshap.timer import StageTimer
from whatshap.cli import log_memory_usage
from whatshap.cli.phase import select_reads, setup_families
from whatshap.cli import CommandLineError, PhasedInputReader


logger = logging.getLogger(__name__)


def int_to_diploid_biallelic_gt(numeric_repr):
    """Converts the classic numeric representation of biallelic, diploid genotypes
    into a genotype object
    """
    if numeric_repr == 0:
        return Genotype([0, 0])
    elif numeric_repr == 1:
        return Genotype([0, 1])
    elif numeric_repr == 2:
        return Genotype([1, 1])
    else:
        return Genotype([])


def determine_genotype(likelihoods: Sequence[float], threshold_prob: float) -> float:
    """given genotype likelihoods for 0/0, 0/1, 1/1, determines likeliest genotype"""

    to_sort = [
        (likelihoods[int_to_diploid_biallelic_gt(0)], 0),
        (likelihoods[int_to_diploid_biallelic_gt(1)], 1),
        (likelihoods[int_to_diploid_biallelic_gt(2)], 2),
    ]
    to_sort.sort(key=lambda x: x[0])

    # make sure there is a unique maximum which is greater than the threshold
    if (to_sort[2][0] > to_sort[1][0]) and (to_sort[2][0] > threshold_prob):
        return int_to_diploid_biallelic_gt(to_sort[2][1])
    else:
        return int_to_diploid_biallelic_gt(-1)


def run_genotype(
    phase_input_files,
    variant_file,
    reference=None,
    output=sys.stdout,
    samples=None,
    chromosomes=None,
    ignore_read_groups=False,
    indels=True,
    mapping_quality=20,
    max_coverage=15,
    nopriors=False,
    ped=None,
    recombrate=1.26,
    genmap=None,
    gt_qual_threshold=0,
    prioroutput=None,
    constant=0.0,
    overhang=10,
    affine_gap=False,
    gap_start=10,
    gap_extend=7,
    mismatch=15,
    write_command_line_header=True,
    use_ped_samples=False,
):
    """
    For now: this function only runs the genotyping algorithm. Genotype likelihoods for
    all variants are computed using the forward backward algorithm
    """
    timers = StageTimer()
    logger.info(
        "This is WhatsHap (genotyping) %s running under Python %s",
        __version__,
        platform.python_version(),
    )
    if write_command_line_header:
        command_line = "(whatshap {}) {}".format(__version__, " ".join(sys.argv[1:]))
    else:
        command_line = None
    with ExitStack() as stack:
        # read the given input files (BAMs, VCFs, ref...)
        numeric_sample_ids = NumericSampleIds()
        phased_input_reader = stack.enter_context(
            PhasedInputReader(
                phase_input_files,
                reference,
                numeric_sample_ids,
                ignore_read_groups,
                indels=indels,
                mapq_threshold=mapping_quality,
                overhang=overhang,
                affine=affine_gap,
                gap_start=gap_start,
                gap_extend=gap_extend,
                default_mismatch=mismatch,
            )
        )
        show_phase_vcfs = phased_input_reader.has_vcfs

        # vcf writer for final genotype likelihoods
        vcf_writer = stack.enter_context(
            GenotypeVcfWriter(command_line=command_line, in_path=variant_file, out_file=output)
        )
        # vcf writer for only the prior likelihoods (if output is desired)
        prior_vcf_writer = None
        if prioroutput is not None:
            prior_vcf_writer = stack.enter_context(
                GenotypeVcfWriter(
                    command_line=command_line,
                    in_path=variant_file,
                    out_file=stack.enter_context(open(prioroutput, "w")),
                )
            )

        # parse vcf with input variants
        # remove all likelihoods that may already be present
        vcf_reader = stack.enter_context(
            VcfReader(
                variant_file, indels=indels, genotype_likelihoods=False, ignore_genotypes=True
            )
        )

        if ignore_read_groups and not samples and len(vcf_reader.samples) > 1:
            raise CommandLineError(
                "When using --ignore-read-groups on a VCF with "
                "multiple samples, --sample must also be used."
            )
        if not samples:
            samples = vcf_reader.samples

        # if --use-ped-samples is set, use only samples from PED file
        if ped and use_ped_samples:
            samples = set()
            for trio in PedReader(ped):
                if trio.child is None or trio.mother is None or trio.father is None:
                    continue
                samples.add(trio.mother)
                samples.add(trio.father)
                samples.add(trio.child)

        vcf_sample_set = set(vcf_reader.samples)
        for sample in samples:
            if sample not in vcf_sample_set:
                raise CommandLineError(
                    f"Sample {sample!r} requested on command-line not found in VCF"
                )

        if ped and genmap:
            logger.info("Using region-specific recombination rates from genetic map %s.", genmap)
            recombination_cost_computer = GeneticMapRecombinationCostComputer(genmap)
        else:
            if ped:
                logger.info("Using uniform recombination rate of %g cM/Mb.", recombrate)
            recombination_cost_computer = UniformRecombinationCostComputer(recombrate)

        samples = frozenset(samples)
        families, family_trios = setup_families(samples, ped, max_coverage)
        for trios in family_trios.values():
            for trio in trios:
                # Ensure that all mentioned individuals have a numeric id
                _ = numeric_sample_ids[trio.child]

        # Read phase information provided as VCF files, if provided.
        with timers("parse_phasing_vcfs"):
            phased_input_reader.read_vcfs()

        # compute genotype likelihood threshold
        gt_prob = 1.0 - (10 ** (-gt_qual_threshold / 10.0))

        for variant_table in timers.iterate("parse_vcf", vcf_reader):

            # create a mapping of genome positions to indices
            var_to_pos = dict()
            for i in range(len(variant_table.variants)):
                var_to_pos[variant_table.variants[i].position] = i

            chromosome = variant_table.chromosome
            if (not chromosomes) or (chromosome in chromosomes):
                logger.info("======== Working on chromosome %r", chromosome)
            else:
                logger.info(
                    "Leaving chromosome %r unchanged (present in VCF but not requested by option --chromosome)",
                    chromosome,
                )
                vcf_writer.write_unchanged(chromosome)
                if prioroutput is not None:
                    prior_vcf_writer.write_unchanged(chromosome)
                continue

            positions = [v.position for v in variant_table.variants]
            if not nopriors:
                # compute prior genotype likelihoods based on all reads
                for sample in samples:
                    logger.info("---- Initial genotyping of %s", sample)
                    with timers("read_bam"):
                        readset, vcf_source_ids = phased_input_reader.read(
                            chromosome, variant_table.variants, sample, read_vcf=False
                        )
                        readset.sort()
                        genotypes, genotype_likelihoods = compute_genotypes(readset, positions)
                        # recompute genotypes based on given threshold
                        reg_genotype_likelihoods = []
                        for gl in range(len(genotype_likelihoods)):
                            norm_sum = (
                                genotype_likelihoods[gl][0]
                                + genotype_likelihoods[gl][1]
                                + genotype_likelihoods[gl][2]
                                + 3 * constant
                            )
                            regularized = PhredGenotypeLikelihoods(
                                [
                                    (genotype_likelihoods[gl][0] + constant) / norm_sum,
                                    (genotype_likelihoods[gl][1] + constant) / norm_sum,
                                    (genotype_likelihoods[gl][2] + constant) / norm_sum,
                                ]
                            )
                            genotypes[gl] = determine_genotype(regularized, gt_prob)
                            assert isinstance(genotypes[gl], Genotype)
                            reg_genotype_likelihoods.append(regularized)
                        variant_table.set_genotype_likelihoods_of(
                            sample,
                            [PhredGenotypeLikelihoods(list(gl)) for gl in reg_genotype_likelihoods],
                        )
                        variant_table.set_genotypes_of(sample, genotypes)
            else:

                # use uniform genotype likelihoods for all individuals
                for sample in samples:
                    variant_table.set_genotype_likelihoods_of(
                        sample, [PhredGenotypeLikelihoods([1 / 3, 1 / 3, 1 / 3])] * len(positions)
                    )

            # if desired, output the priors in separate vcf
            if prioroutput is not None:
                prior_vcf_writer.write_genotypes(chromosome, variant_table, indels)

            # Iterate over all families to process, i.e. a separate DP table is created
            # for each family.
            for representative_sample, family in sorted(families.items()):
                if len(family) == 1:
                    logger.info("---- Processing individual %s", representative_sample)
                else:
                    logger.info("---- Processing family with individuals: %s", ",".join(family))
                max_coverage_per_sample = max(1, max_coverage // len(family))
                logger.info("Using maximum coverage per sample of %dX", max_coverage_per_sample)
                trios = family_trios[representative_sample]
                assert (len(family) == 1) or (len(trios) > 0)

                # Get the reads belonging to each sample
                readsets = dict()
                for sample in family:
                    with timers("read_bam"):
                        readset, vcf_source_ids = phased_input_reader.read(
                            chromosome, variant_table.variants, sample
                        )

                    with timers("select"):
                        readset = readset.subset(
                            [i for i, read in enumerate(readset) if len(read) >= 2]
                        )
                        logger.info(
                            "Kept %d reads that cover at least two variants each", len(readset)
                        )
                        selected_reads = select_reads(
                            readset, max_coverage_per_sample, preferred_source_ids=vcf_source_ids
                        )
                    readsets[sample] = selected_reads

                # Merge reads into one ReadSet (note that each Read object
                # knows the sample it originated from).
                all_reads = ReadSet()
                for sample, readset in readsets.items():
                    for read in readset:
                        assert read.is_sorted(), "Add a read.sort() here"
                        all_reads.add(read)

                all_reads.sort()

                # Determine which variants can (in principle) be phased
                accessible_positions = sorted(all_reads.get_positions())
                logger.info(
                    "Variants covered by at least one phase-informative "
                    "read in at least one individual after read selection: %d",
                    len(accessible_positions),
                )

                # Create Pedigree
                pedigree = Pedigree(numeric_sample_ids)
                for sample in family:
                    # genotypes are assumed to be unknown, so ignore information that
                    # might already be present in the input vcf
                    all_genotype_likelihoods = variant_table.genotype_likelihoods_of(sample)
                    genotype_l = [
                        all_genotype_likelihoods[var_to_pos[a_p]] for a_p in accessible_positions
                    ]
                    pedigree.add_individual(
                        sample, [Genotype([]) for i in range(len(accessible_positions))], genotype_l
                    )
                for trio in trios:
                    pedigree.add_relationship(
                        father_id=trio.father, mother_id=trio.mother, child_id=trio.child
                    )

                recombination_costs = recombination_cost_computer.compute(accessible_positions)

                # Finally, run genotyping algorithm
                with timers("genotyping"):
                    problem_name = "genotyping"
                    logger.info(
                        "Genotype %d sample%s by solving the %s problem ...",
                        len(family),
                        "s" if len(family) > 1 else "",
                        problem_name,
                    )
                    forward_backward_table = GenotypeDPTable(
                        numeric_sample_ids,
                        all_reads,
                        recombination_costs,
                        pedigree,
                        accessible_positions,
                    )
                    # store results
                    for s in family:
                        likelihood_list = variant_table.genotype_likelihoods_of(s)
                        genotypes_list = variant_table.genotypes_of(s)

                        for pos in range(len(accessible_positions)):
                            likelihoods = forward_backward_table.get_genotype_likelihoods(s, pos)

                            # compute genotypes from likelihoods and store information
                            geno = determine_genotype(likelihoods, gt_prob)
                            assert isinstance(geno, Genotype)
                            genotypes_list[var_to_pos[accessible_positions[pos]]] = geno
                            likelihood_list[var_to_pos[accessible_positions[pos]]] = likelihoods

                        variant_table.set_genotypes_of(s, genotypes_list)
                        variant_table.set_genotype_likelihoods_of(s, likelihood_list)

            with timers("write_vcf"):
                logger.info("======== Writing VCF")
                vcf_writer.write_genotypes(chromosome, variant_table, indels)
                logger.info("Done writing VCF")

            logger.debug("Chromosome %r finished", chromosome)

    logger.info("\n== SUMMARY ==")
    total_time = timers.total()
    log_memory_usage()
    logger.info("Time spent reading BAM:                      %6.1f s", timers.elapsed("read_bam"))
    logger.info("Time spent parsing VCF:                      %6.1f s", timers.elapsed("parse_vcf"))
    if show_phase_vcfs:
        logger.info(
            "Time spent parsing input phasings from VCFs: %6.1f s",
            timers.elapsed("parse_phasing_vcfs"),
        )
    logger.info("Time spent selecting reads:                  %6.1f s", timers.elapsed("select"))
    logger.info(
        "Time spent genotyping:                          %6.1f s", timers.elapsed("genotyping")
    )
    logger.info("Time spent writing VCF:                      %6.1f s", timers.elapsed("write_vcf"))
    logger.info("Time spent on rest:                          %6.1f s", total_time - timers.sum())
    logger.info("Total elapsed time:                          %6.1f s", total_time)


# fmt: off
def add_arguments(parser):
    arg = parser.add_argument
    # Positional arguments
    arg('variant_file', metavar='VCF', help='VCF file with variants to be genotyped (can be gzip-compressed)')
    arg('phase_input_files', nargs='*', metavar='PHASEINPUT',
        help='BAM or VCF file(s) with phase information, either through sequencing reads (BAM) or through phased blocks (VCF)')

    arg('-o', '--output', default=sys.stdout,
        help='Output VCF file. Add .gz to the file name to get compressed output. '
        'If omitted, use standard output.')
    arg('--reference', '-r', metavar='FASTA',
        help='Reference file. Provide this to detect alleles through re-alignment. '
        'If no index (.fai) exists, it will be created')

    arg = parser.add_argument_group('Input pre-processing, selection and filtering').add_argument
    arg('--max-coverage', '-H', metavar='MAXCOV', default=15, type=int,
        help='Reduce coverage to at most MAXCOV (default: %(default)s).')
    arg('--mapping-quality', '--mapq', metavar='QUAL',
        default=20, type=int, help='Minimum mapping quality (default: %(default)s)')
    arg('--indels', dest='indels', default=False, action='store_true',
        help='Also genotype indels (default: genotype only SNPs)')
    arg('--ignore-read-groups', default=False, action='store_true',
        help='Ignore read groups in BAM header and assume all reads come '
        'from the same sample.')
    arg('--sample', dest='samples', metavar='SAMPLE', default=[], action='append',
        help='Name of a sample to genotype. If not given, all samples in the '
        'input VCF are genotyped. Can be used multiple times.')
    arg('--chromosome', dest='chromosomes', metavar='CHROMOSOME', default=[], action='append',
        help='Name of chromosome to genotyped. If not given, all chromosomes in the '
        'input VCF are genotyped. Can be used multiple times.')
    arg('--gt-qual-threshold', metavar='GTQUALTHRESHOLD', type=float, default=0,
        help='Phred scaled error probability threshold used for genotyping (default: %(default)s). Must be at least 0. '
        'If error probability of genotype is higher, genotype ./. is output.')
    arg('--no-priors', dest='nopriors', default=False, action='store_true',
        help='Skip initial prior genotyping and use uniform priors (default: %(default)s).')
    arg('-p', '--prioroutput', default=None,
        help='output prior genotype likelihoods to the given file.')
    arg('--overhang', metavar='OVERHANG', default=10, type=int,
        help='When --reference is used, extend alignment by this many bases to left and right when realigning (default: %(default)s).')
    arg('--constant', metavar='CONSTANT', default=0, type=float,
        help='This constant is used to regularize the priors (default: %(default)s).')
    arg('--affine-gap', default=False, action='store_true',
        help='When detecting alleles through re-alignment, use affine gap costs (EXPERIMENTAL).')
    arg('--gap-start', metavar='GAPSTART', default=10, type=float,
        help='gap starting penalty in case affine gap costs are used (default: %(default)s).')
    arg('--gap-extend', metavar='GAPEXTEND', default=7, type=float,
        help='gap extend penalty in case affine gap costs are used (default: %(default)s).')
    arg('--mismatch', metavar='MISMATCH', default=15, type=float,
        help='mismatch cost in case affine gap costs are used (default: %(default)s)')

    arg = parser.add_argument_group('Pedigree genotyping').add_argument
    arg('--ped', metavar='PED/FAM',
        help='Use pedigree information in PED file to improve genotyping '
        '(switches to PedMEC algorithm). Columns 2, 3, 4 must refer to child, '
        'mother, and father sample names as used in the VCF and BAM. Other '
        'columns are ignored (EXPERIMENTAL).')
    arg('--recombrate', metavar='RECOMBRATE', type=float, default=1.26,
        help='Recombination rate in cM/Mb (used with --ped). If given, a constant recombination '
        'rate is assumed (default: %(default)gcM/Mb).')
    arg('--genmap', metavar='FILE',
        help='File with genetic map (used with --ped) to be used instead of constant recombination '
        'rate, i.e. overrides option --recombrate.')
    arg('--use-ped-samples', dest='use_ped_samples',
        action='store_true', default=False,
        help='Only work on samples mentioned in the provided PED file.')
# fmt: on


def validate(args, parser):
    if args.ignore_read_groups and args.ped:
        parser.error("Option --ignore-read-groups cannot be used together with --ped")
    if args.genmap and not args.ped:
        parser.error("Option --genmap can only be used together with --ped")
    if args.genmap and len(args.chromosomes) != 1:
        parser.error(
            "Option --genmap can only be used when working on exactly one chromosome (use --chromosome)"
        )
    if len(args.phase_input_files) == 0:
        parser.error("Not providing any PHASEINPUT files not allowed for genotyping.")
    if args.gt_qual_threshold < 0:
        parser.error("Genotype quality threshold (gt-qual-threshold) must be at least 0.")
    if args.prioroutput is not None and args.nopriors:
        parser.error("Genotype priors are only computed if --no-priors is NOT set.")
    if args.constant != 0 and args.nopriors:
        parser.error("--constant can only be used if --no-priors is NOT set..")
    if args.affine_gap and not args.reference:
        parser.error("Option --affine-gap can only be used together with --reference.")
    if args.use_ped_samples and not args.ped:
        parser.error("Option --use-ped-samples can only be used when PED file is provided (--ped).")
    if args.use_ped_samples and args.samples:
        parser.error("Option --use-ped-samples cannot be used together with --samples")


def main(args):
    run_genotype(**vars(args))



from typing import List, Tuple, Iterator, Dict, Sequence, Optional, Iterable

from whatshap.core import ReadSet

class ClusterEditingSolver:
    def __init__(self, m: TriangleSparseMatrix, bundle_edges: bool): ...
    def run(self) -> List[List[int]]: ...

class TriangleSparseMatrix:
    def __init__(self): ...
    def get(self, i: int, j: int) -> float: ...
    def set(self, i: int, j: int, value: float) -> None: ...
    def size(self) -> int: ...
    def getEntries(self) -> List[Tuple[int, int]]: ...
    def __iter__(self) -> Iterator[Tuple[int, int]]: ...
    def __len__(self) -> int: ...

# class ReadScoring:
#     def __init__(self): ...
def scoreReadset(
    readset: ReadSet, min_overlap: int, ploidy: int, err: float = ...
) -> TriangleSparseMatrix: ...

class HaploThreader:
    def __init__(
        self,
        ploidy: int,
        switch_cost: float,
        affine_switch_cost: float,
        symmetry_optimization: bool,
        row_limit: int,
    ): ...
    def computePathsBlockwise(
        self,
        block_starts: List[int],
        coverage_map: List[List[int]],
        coverage: List[List[float]],
        consensus: List[List[int]],
        genotypes: List[Dict[int, int]],
    ) -> List[List[int]]: ...
    def computePaths(
        self,
        start: int,
        end: int,
        coverage_map: List[List[int]],
        coverage: List[List[float]],
        consensus: List[List[int]],
        genotypes: List[Dict[int, int]],
    ) -> List[List[int]]: ...

class SwitchFlipCalculator:
    def __init__(self, ploidy: int, switch_cost: int = ..., flip_cost: int = ...): ...
    def compute_switch_flips_poly(
        self, phasing0: Sequence[str], phasing1: Sequence[str]
    ) -> Tuple[float, float, List[int], List[List[int]], List[List[int]]]: ...

def compute_polyploid_genotypes(
    readset: ReadSet, ploidy: int, positions: Optional[Iterable[int]] = ...
) -> List[List[int]]: ...



# cython: language_level=3

import math
import logging
from collections import defaultdict

from .coverage import CovMonitor
from .graph import ComponentFinder
from .priorityqueue import PriorityQueue

from libcpp.unordered_set cimport unordered_set

from .priorityqueue cimport priority_type, priority_type_ptr, queue_entry_type, PriorityQueue
from .core cimport ReadSet
from . cimport cpp

logger = logging.getLogger(__name__)


def _construct_indexes(readset, preferred_source_ids=None):
	''' The parameter readset: is the given ReadSet and returns, all possible variant positions, the vcf_index_ mapping
    and the variant_to_reads_map'''
	positions = readset.get_positions()
	vcf_indices = {position: index for index, position in enumerate(positions)}
	variant_to_reads_map = defaultdict(list)
	preferred_reads = set()
	for index, read in enumerate(readset):
		if preferred_source_ids is not None:
			if read.source_id in preferred_source_ids:
				preferred_reads.add(index)
		for variant in read:
			variant_index = vcf_indices[variant.position]
			variant_to_reads_map[variant_index].append(index)
	return positions, vcf_indices, variant_to_reads_map, preferred_reads


cdef priority_type_ptr _update_score_for_reads(priority_type_ptr former_score, cpp.ReadSet* readset, index, unordered_set[int]& already_covered_variants):
	'''updatest the score of the read, depending on how many reads are already covered'''
	cdef int first_score = former_score.at(0)
	cdef int second_score = former_score.at(1)
	cdef int quality = former_score.at(2)
	cdef cpp.Read* read = readset.get(index)
	cdef unordered_set[int].iterator it
	for i in range(read.getVariantCount()):
		it = already_covered_variants.find(read.getPosition(i))
		if it == already_covered_variants.end():
			first_score -= 1
	cdef priority_type_ptr result = new priority_type()
	result.push_back(first_score)
	result.push_back(second_score)
	result.push_back(quality)
	return result


cdef priority_type_ptr _compute_score_for_read(cpp.ReadSet* readset, int index, vcf_indices):
	'''Compute the score for one read, assuming no other reads have been selected so far
	(after selecting reads, scores can be updated using _update_score_for_reads).
	We use the following scoring scheme: (new - gaps, total - bad, min(qualities)),
	where "new" is the number of variants covered by this read and no other selected read (so far),
	"gaps" is the number of variants overlapped by (physical) fragment, but not covered by the sequenced part of the read,
	"total" is the total number of variants covered by the read, and "min(qualities)" is the minimum
	over all base qualities at variant positions.
	'''
	cdef cpp.Read* read = readset.get(index)
	cdef int min_quality = -1
	cdef int good_score = 0
	cdef int bad_score = 0
	cdef int quality = -1
	cdef int pos = -1
	covered_variants = []
	for i in range(read.getVariantCount()):
		quality = read.getVariantQuality(i)
		pos = read.getPosition(i)
		if i == 0:
			min_quality = quality
		else:
			min_quality = min(min_quality, quality)
		variant_covered = vcf_indices.get(pos)
		if variant_covered != None:
			covered_variants.append(variant_covered)
			#good_score represents how many variants are really covered
			good_score += 1
	#bad_score here means number of variants which are not covered...
	if len(covered_variants) != (covered_variants[len(covered_variants) - 1] - covered_variants[0] + 1):
		bad_score = (covered_variants[len(covered_variants) - 1] - covered_variants[0] + 1) - len(covered_variants)
	cdef priority_type_ptr result = new priority_type()
	#initially new_score is the same as good_score, but new_score is updated later
	result.push_back(good_score - bad_score)
	result.push_back(good_score - bad_score)
	result.push_back(min_quality)
	return result


cdef PriorityQueue _construct_priorityqueue(cpp.ReadSet* readset, read_indices, vcf_indices):
	'''Construct a priority queue containing all given read indicies, each one representing the
	respective read from readset.'''
	cdef PriorityQueue priorityqueue = PriorityQueue()

	for index in read_indices:
		read = readset.get(index)
		computed_score = _compute_score_for_read(readset, index, vcf_indices)
		priorityqueue.c_push(computed_score, index)

	return priorityqueue


cdef _slice_read_selection(PriorityQueue pq, coverages, max_cov, cpp.ReadSet* readset, vcf_indices, variant_to_reads_map):
	'''Extraction of a set of read indices, where each variant should be covered at least once, if coverage, or reads are allowing it '''
	# positions of variants covered by any read selected so far
	already_covered_variants = set()
	# indices of selected reads
	reads_in_slice = set()
	# indices of reads that cannot be added because doing that would violate coverage constraint
	reads_violating_coverage = set()
	cdef cpp.Read* extracted_read = NULL
	cdef queue_entry_type entry
	cdef priority_type_ptr max_score
	cdef int max_item
	cdef int pos
	cdef unordered_set[int] variants_covered_by_this_read
	cdef priority_type_ptr oldscore
	cdef priority_type_ptr newscore
	while not pq.c_is_empty():
		variants_covered_by_this_read.clear()
		entry = pq.c_pop()
		max_score = entry.first
		max_item = entry.second
		extracted_read = readset.get(max_item)
		covers_new_variant = False
		#look if positions covered by this reads are already covered or not
		for i in range(extracted_read.getVariantCount()):
			pos = extracted_read.getPosition(i)
			if pos in already_covered_variants:
				continue
			else:
				covers_new_variant = True
				#stores the positions the read covers
				variants_covered_by_this_read.insert(pos)
		# only add read if it covers at least one new variant and adding it does not violate coverage constraints
		begin = vcf_indices.get(extracted_read.getPosition(0))
		end = vcf_indices.get(extracted_read.getPosition(extracted_read.getVariantCount() - 1)) + 1
		if coverages.max_coverage_in_range(begin, end) >= max_cov:
			reads_violating_coverage.add(max_item)
		elif covers_new_variant:
			coverages.add_read(begin, end)
			reads_in_slice.add(max_item)
			reads_whose_score_has_to_be_updated = set()

			#again go over the positions in the read and add them to the already_covered_variants list

			#Only the positions in the read which cover new variants are analysed
			for pos in variants_covered_by_this_read:
				already_covered_variants.add(pos)
				reads_whose_score_has_to_be_updated.update(variant_to_reads_map[vcf_indices.get(pos)])

			#find difference between to_decrease_score and selected_reads in order to not to try to decrease score by selected reads
			selected_read_set = set(reads_in_slice)
			decrease_set = reads_whose_score_has_to_be_updated
			d_set = decrease_set.difference(selected_read_set)

			#Catch with None if element is not in the priorityqueue
			for element in d_set:
				oldscore = pq.c_get_score_by_item(element)
				if oldscore != NULL:
					newscore = _update_score_for_reads(oldscore, readset, element, variants_covered_by_this_read)
					pq.c_change_score(element, newscore)
	return reads_in_slice, reads_violating_coverage



cdef format_read_source_stats(cpp.ReadSet* readset, indices):
	"""Creates a string giving information on the source_ids of the reads with the given indices."""
	if len(indices) == 0:
		return 'n/a'
	source_id_counts = defaultdict(int)
	for i in indices:
		read = readset.get(i)
		source_id_counts[read.getSourceID()] += 1
	present_ids = list(source_id_counts.keys())
	present_ids.sort()
	return ', '.join('{}:{}'.format(source_id, count) for source_id, count in source_id_counts.items())


cdef readselection_helper(coverages, max_cov, cpp.ReadSet* readset, vcf_indices, variant_to_reads_map, selected_reads, undecided_reads, positions, bridging):
	cdef PriorityQueue pq
	cdef cpp.Read* read
	loop = 0
	while len(undecided_reads) > 0:
		pq = _construct_priorityqueue(readset, undecided_reads, vcf_indices)
		reads_in_slice, reads_violating_coverage = _slice_read_selection(pq, coverages, max_cov, readset, vcf_indices, variant_to_reads_map)
		selected_reads.update(reads_in_slice)
		undecided_reads -= reads_in_slice
		undecided_reads -= reads_violating_coverage

		# Create new component finder from reads just selected
		component_finder = ComponentFinder(positions)
		for read_index in reads_in_slice:
			read = readset.get(read_index)
			for i in range(1, read.getVariantCount()):
				component_finder.merge(read.getPosition(0), read.getPosition(i))

		bridging_reads = set()
		if bridging:
			pq = _construct_priorityqueue(readset, undecided_reads, vcf_indices)
			while not pq.is_empty():
				score, read_index = pq.pop()
				read = readset.get(read_index)
				covered_blocks = set()
				for i in range(read.getVariantCount()):
					covered_blocks.add(component_finder.find(read.getPosition(i)))

				# check whether read meets coverage constraints
				begin = vcf_indices.get(read.getPosition(0))
				end = vcf_indices.get(read.getPosition(read.getVariantCount() - 1)) + 1
				if coverages.max_coverage_in_range(begin, end) >= max_cov:
					undecided_reads.remove(read_index)
					continue

				# skip read if it only covers one block
				if len(covered_blocks) < 2:
					continue
				bridging_reads.add(read_index)
				selected_reads.add(read_index)

				coverages.add_read(begin, end)

				undecided_reads.remove(read_index)
				# update component_finder
				for i in range(1, read.getVariantCount()):
					component_finder.merge(read.getPosition(0), read.getPosition(i))
		loop += 1
		logger.debug(
			'... iteration %d: selected %d reads (source: %s) to cover positions and %d reads (source: %s) for bridging; %d reads left undecided',
			loop, len(reads_in_slice), format_read_source_stats(readset, reads_in_slice), len(bridging_reads),
			format_read_source_stats(readset, bridging_reads), len(undecided_reads)
		)
	return selected_reads


def readselection(ReadSet pyreadset, max_cov, preferred_source_ids=None, bridging=True):
	'''Return the selected readindices which do not violate the maximal coverage, and additionally usage of a boolean for deciding if
     the bridging is needed or not.'''

	cdef cpp.ReadSet* readset = pyreadset.thisptr
	assert readset != NULL

	positions, vcf_indices, variant_to_reads_map, preferred_reads = _construct_indexes(pyreadset, preferred_source_ids)

	logger.debug('Running read selection for %d reads covering %d variants (bridging %s)', len(pyreadset), len(positions),'ON' if bridging else 'OFF')

	#initialization of Coverage Monitor
	coverages = CovMonitor(len(positions))

	# indices of reads that have been selected
	selected_reads = set()

	for r in pyreadset:
		if not len(r) >= 2:
			print(r)
			raise ValueError('readselection expects reads that cover at least two variants')

	# indices of reads that could (potentially) still be selected
	undecided_reads = set(range(len(pyreadset)))
	
	if len(preferred_reads) > 0:
		selected_preferred_reads = readselection_helper(coverages, max_cov, readset, vcf_indices, variant_to_reads_map, selected_reads, preferred_reads, positions, bridging)
		selected_reads.update(selected_preferred_reads)
		undecided_reads -= preferred_reads
	
	selected_reads = readselection_helper(coverages, max_cov, readset, vcf_indices, variant_to_reads_map, selected_reads, undecided_reads, positions, bridging)

	return selected_reads



import math

from pytest import raises, approx, fixture
from whatshap.core import PhredGenotypeLikelihoods, Genotype
from whatshap.cli.phase import run_whatshap
from whatshap.vcf import (
    VcfReader,
    MixedPhasingError,
    PloidyError,
    VariantCallPhase,
    VcfVariant,
    GenotypeLikelihoods,
    VcfIndexMissing,
)
from whatshap.testhelpers import (
    canonic_index_to_biallelic_gt,
    canonic_index_list_to_biallelic_gt_list,
)

import pysam


@fixture(params=["whatshap", "hapchat"])
def algorithm(request):
    return request.param


def test_read_phased():
    tables = list(VcfReader("tests/data/phasedinput.vcf", phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "ref"
    assert table.samples == ["sample"]
    assert len(table.variants) == 2
    assert table.variants[0].reference_allele == "A"
    assert table.variants[0].alternative_allele == "C"
    assert table.variants[1].reference_allele == "G"
    assert table.variants[1].alternative_allele == "T"
    assert table.genotypes[0][0] == table.genotypes[0][1] == canonic_index_to_biallelic_gt(1)


def test_read_multisample_vcf():
    tables = list(VcfReader("tests/data/multisample.vcf"))
    assert len(tables) == 2
    table, table_b = tables
    assert table_b.chromosome == "chrB"
    assert table_b.samples == ["sample1", "sample2"]

    assert table.chromosome == "chrA"
    assert len(table.variants) == 3
    assert table.samples == ["sample1", "sample2"]

    assert table.variants[0].reference_allele == "A"
    assert table.variants[0].alternative_allele == "T"
    assert table.variants[1].reference_allele == "C"
    assert table.variants[1].alternative_allele == "G"
    assert table.variants[2].reference_allele == "G"
    assert table.variants[2].alternative_allele == "T"

    assert len(table.genotypes) == 2
    assert list(table.genotypes[0]) == canonic_index_list_to_biallelic_gt_list([1, 1, 1])
    assert list(table.genotypes[1]) == canonic_index_list_to_biallelic_gt_list([1, 1, 0])

    assert list(table.genotypes_of("sample1")) == canonic_index_list_to_biallelic_gt_list([1, 1, 1])
    assert list(table.genotypes_of("sample2")) == canonic_index_list_to_biallelic_gt_list([1, 1, 0])


def test_read_phased_vcf():
    for filename in ["tests/data/phased-via-HP.vcf", "tests/data/phased-via-PS.vcf"]:
        print("Testing", filename)
        tables = list(VcfReader(filename, phases=True))
        assert len(tables) == 2
        table_a, table_b = tables

        assert table_a.chromosome == "chrA"
        assert len(table_a.variants) == 4
        assert table_a.samples == ["sample1", "sample2"]

        assert table_b.chromosome == "chrB"
        assert len(table_b.variants) == 2
        assert table_b.samples == ["sample1", "sample2"]

        assert len(table_a.genotypes) == 2
        assert list(table_a.genotypes[0]) == canonic_index_list_to_biallelic_gt_list([1, 2, 1, 1])
        assert list(table_a.genotypes[1]) == canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1])
        assert list(table_a.genotypes_of("sample1")) == canonic_index_list_to_biallelic_gt_list(
            [1, 2, 1, 1]
        )
        assert list(table_a.genotypes_of("sample2")) == canonic_index_list_to_biallelic_gt_list(
            [1, 1, 1, 1]
        )

        assert len(table_b.genotypes) == 2
        assert list(table_b.genotypes[0]) == canonic_index_list_to_biallelic_gt_list([0, 1])
        assert list(table_b.genotypes[1]) == canonic_index_list_to_biallelic_gt_list([1, 2])
        assert list(table_b.genotypes_of("sample1")) == canonic_index_list_to_biallelic_gt_list(
            [0, 1]
        )
        assert list(table_b.genotypes_of("sample2")) == canonic_index_list_to_biallelic_gt_list(
            [1, 2]
        )

        print(table_a.phases)
        assert len(table_a.phases) == 2
        expected_phase_sample1 = [
            None,
            None,
            VariantCallPhase(block_id=300, phase=(1, 0), quality=23),
            VariantCallPhase(block_id=300, phase=(0, 1), quality=42),
        ]
        expected_phase_sample2 = [
            VariantCallPhase(block_id=100, phase=(0, 1), quality=10),
            VariantCallPhase(block_id=100, phase=(1, 0), quality=20),
            VariantCallPhase(block_id=300, phase=(0, 1), quality=30),
            VariantCallPhase(block_id=300, phase=(0, 1), quality=None),
        ]
        assert list(table_a.phases[0]) == expected_phase_sample1
        assert list(table_a.phases[1]) == expected_phase_sample2
        assert list(table_a.phases_of("sample1")) == expected_phase_sample1
        assert list(table_a.phases_of("sample2")) == expected_phase_sample2

        assert len(table_b.phases) == 2
        assert list(table_b.phases[0]) == [None, None]
        assert list(table_b.phases[1]) == [None, None]
        assert list(table_b.phases_of("sample1")) == [None, None]
        assert list(table_b.phases_of("sample2")) == [None, None]


def test_mixed_phasing_vcf():
    with raises(MixedPhasingError):
        list(VcfReader("tests/data/phased-via-mixed-HP-PS.vcf", phases=True))


def test_vcf_variant_hashability():
    v = [
        VcfVariant(10, "A", "TC"),
        VcfVariant(10, "A", "TCA"),
        VcfVariant(10, "C", "TC"),
        VcfVariant(20, "A", "TC"),
        VcfVariant(10, "A", "TCA"),
        VcfVariant(20, "A", "TC"),
    ]
    assert len(set(v)) == 4


def test_phasing_to_reads():
    for filename in ["tests/data/phased-via-HP.vcf", "tests/data/phased-via-PS.vcf"]:
        tables = list(VcfReader(filename, phases=True))
        assert len(tables) == 2
        table_a, table_b = tables
        phase_reads_sample1 = list(
            table_a.phased_blocks_as_reads(
                "sample1", table_a.variants, 17, 18, default_quality=90, mapq=101
            )
        )
        print(phase_reads_sample1)
        assert len(phase_reads_sample1) == 1
        read = phase_reads_sample1[0]
        assert len(read) == 2
        assert read.name == "sample1_block_300"
        assert read.source_id == 17
        assert read.mapqs == (101,)
        assert read[0].position == 300 - 1
        assert read[0].allele == 1
        assert read[0].quality == 23
        assert read[1].position == 350 - 1
        assert read[1].allele == 0
        assert read[1].quality == 42

        phase_reads_sample2 = list(
            table_a.phased_blocks_as_reads(
                "sample2", table_a.variants, 11, 12, default_quality=91, mapq=102
            )
        )
        print(phase_reads_sample2)
        assert len(phase_reads_sample2) == 2
        read1, read2 = phase_reads_sample2
        assert len(read1) == len(read2) == 2
        if read1[0].position > read2[0].position:
            read1, read2 = read2, read1
        assert read1.name == "sample2_block_100"
        assert read1.source_id == 11
        assert read1.mapqs == (102,)
        assert read1[0].position == 100 - 1
        assert read1[0].allele == 0
        assert read1[0].quality == 10
        assert read1[1].position == 150 - 1
        assert read1[1].allele == 1
        assert read1[1].quality == 20
        assert read2.name == "sample2_block_300"
        assert read2.source_id == 11
        assert read2.mapqs == (102,)
        assert read2[0].position == 300 - 1
        assert read2[0].allele == 0
        assert read2[0].quality == 30
        assert read2[1].position == 350 - 1
        assert read2[1].allele == 0
        assert read2[1].quality == 91

        variants = [
            VcfVariant(350 - 1, "G", "T"),
            VcfVariant(300 - 1, "G", "T"),
            VcfVariant(17, "A", "TTC"),
            VcfVariant(1000, "C", "G"),
        ]
        phase_reads_sample2 = list(
            table_a.phased_blocks_as_reads(
                "sample2", variants, 11, 12, default_quality=91, mapq=102
            )
        )
        print(phase_reads_sample2)
        assert len(phase_reads_sample2) == 1
        read = phase_reads_sample2[0]
        assert len(read) == 2
        assert read.name == "sample2_block_300"
        assert read.source_id == 11
        assert read.mapqs == (102,)
        assert read[0].position == 300 - 1
        assert read[0].allele == 0
        assert read[0].quality == 30
        assert read[1].position == 350 - 1
        assert read[1].allele == 0
        assert read[1].quality == 91


def test_unknown_genotype():
    """VCF with './.' genotype"""
    tables = list(VcfReader("tests/data/unknown-genotype.vcf"))
    assert tables[0].genotypes[1][0] == Genotype([])
    assert tables[0].genotypes[1][0].is_none()


def test_normalize():
    assert VcfVariant(100, "A", "C").normalized() == VcfVariant(100, "A", "C")
    assert VcfVariant(100, "", "A").normalized() == VcfVariant(100, "", "A")
    assert VcfVariant(100, "A", "").normalized() == VcfVariant(100, "A", "")
    assert VcfVariant(100, "A", "AC").normalized() == VcfVariant(101, "", "C")
    assert VcfVariant(100, "AC", "A").normalized() == VcfVariant(101, "C", "")
    assert VcfVariant(100, "ACAGACC", "ACAGACT").normalized() == VcfVariant(106, "C", "T")
    assert VcfVariant(100, "GCTG", "GCTAAA").normalized() == VcfVariant(103, "G", "AAA")
    assert VcfVariant(100, "ATTA", "ATA").normalized() == VcfVariant(101, "T", "")
    assert VcfVariant(100, "ATTTC", "ATTTTTTC").normalized() == VcfVariant(101, "", "TTT")
    assert VcfVariant(100, "GCTGTT", "GCTAAATT").normalized() == VcfVariant(103, "G", "AAA")


def test_read_duplicate_position():
    """Two rows with same position"""
    # As soon as we can actually work with multiple such rows, this test
    # needs to be updated since it currently just checks whether the second of
    # the positions is skipped.
    table = list(VcfReader("tests/data/duplicate-positions.vcf", indels=True))[0]
    assert len(table.variants) == 2
    assert table.variants[0].position == 1
    assert table.variants[0].reference_allele == "A"
    assert table.variants[0].alternative_allele == "T"
    assert table.variants[1].position == 19
    assert table.variants[1].reference_allele == "G"
    assert table.variants[1].alternative_allele == "A"


def test_do_not_phase_duplicate_position(algorithm, tmpdir):
    """Ensure HP tag is added only to first of duplicate positions"""
    tmpvcf = str(tmpdir.join("duplicate-positions-phased.vcf"))
    run_whatshap(
        phase_input_files=["tests/data/oneread.bam"],
        variant_file="tests/data/duplicate-positions.vcf",
        output=tmpvcf,
        algorithm=algorithm,
    )
    seen_positions = set()
    records = list(pysam.VariantFile(tmpvcf))
    assert len(records) == 4
    for record in records:
        assert not (record.start in seen_positions and "HP" in record.format)
        seen_positions.add(record.start)


def test_multi_alt():
    """Skip multi-ALT in VCF"""
    table = list(VcfReader("tests/data/unknown-genotype.vcf"))[0]
    assert [variant.position for variant in table.variants] == [1, 4]


def assert_genotype_likelihoods(actual, expected):
    if expected is None:
        assert actual is None
        return
    for i in range(2):
        e = expected.log10_prob_of(i)
        a = actual.log10_prob_of(i)
        if e is None or a is None:
            assert a is None and e is None
        else:
            assert e == approx(a, rel=1e-6)


def test_read_genotype_likelihoods():
    tables = list(VcfReader("tests/data/genotype-likelihoods.vcf", genotype_likelihoods=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "chrA"
    assert table.samples == ["sample1", "sample2"]
    assert len(table.variants) == 4

    assert len(table.genotypes) == 2
    assert list(table.genotypes[0]) == canonic_index_list_to_biallelic_gt_list([2, 1, 1, 1])
    assert list(table.genotypes[1]) == canonic_index_list_to_biallelic_gt_list([1, 0, 0, 1])

    gl0 = GenotypeLikelihoods([-2.1206, -0.8195, -0.07525])
    gl1 = GenotypeLikelihoods([-10.3849, 0, -5.99143])
    gl2 = GenotypeLikelihoods([-2.1, None, -0.8])
    gl3 = GenotypeLikelihoods([0, -10.0, -0.6])

    assert len(table.genotype_likelihoods_of("sample1")) == 4
    assert len(table.genotype_likelihoods_of("sample2")) == 4

    expected1 = [gl0, gl2, None, gl0]
    expected2 = [gl1, gl3, None, gl1]
    for actual_gl, expected_gl in zip(table.genotype_likelihoods_of("sample1"), expected1):
        assert_genotype_likelihoods(actual_gl, expected_gl)
    for actual_gl, expected_gl in zip(table.genotype_likelihoods_of("sample2"), expected2):
        assert_genotype_likelihoods(actual_gl, expected_gl)


def test_genotype_likelihoods():
    assert list(PhredGenotypeLikelihoods([0, 0, 0])) == [0, 0, 0]
    assert list(PhredGenotypeLikelihoods([7, 1, 12])) == [7, 1, 12]
    gl = GenotypeLikelihoods([math.log10(x) for x in [1e-10, 0.5, 0.002]])
    assert list(gl.as_phred()) == [97, 0, 24]
    assert list(gl.as_phred(regularizer=0.01)) == [20, 0, 19]


def test_read_region():
    vcf_reader = VcfReader("tests/data/haplotag_1.vcf.gz")
    tableA = vcf_reader.fetch("chr1")
    tableB = vcf_reader.fetch("chr1", 1_069_570, 1_080_000)
    assert tableA.chromosome == tableB.chromosome
    assert len(tableA.variants) == len(tableB.variants)


def test_read_region_subsets():
    regions = [(1069570, 1070690), (1074910, 1076152)]
    vcf_reader = VcfReader("tests/data/haplotag_1.vcf.gz", indels=True)
    table = vcf_reader.fetch_regions("chr1", regions)
    assert table.chromosome == "chr1"
    assert len(table.variants) == 8
    assert table.variants[5].reference_allele == "CG"
    assert table.variants[5].alternative_allele == "C"


def test_read_tetraploid_unphased():
    tables = list(VcfReader("tests/data/polyploid.chr22.unphased.vcf", phases=False))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "chr22"
    assert table.samples == ["HG00514_NA19240"]
    assert len(table.variants) == 8
    assert table.variants[0].reference_allele == "A"
    assert table.variants[0].alternative_allele == "C"
    assert table.variants[1].reference_allele == "G"
    assert table.variants[1].alternative_allele == "A"
    assert table.variants[2].reference_allele == "G"
    assert table.variants[2].alternative_allele == "T"
    assert table.variants[3].reference_allele == "G"
    assert table.variants[3].alternative_allele == "C"
    print("Got:")
    for genotype in table.genotypes[0]:
        print(genotype)
    print("Exp:")
    for genotype in canonic_index_list_to_biallelic_gt_list([3, 2, 0, 3, 3, 1, 1, 1]):
        print(genotype)
    assert table.genotypes[0] == canonic_index_list_to_biallelic_gt_list(
        [3, 2, 0, 3, 3, 1, 1, 1], 4
    )


def test_read_tetraploid_phased():
    tables = list(VcfReader("tests/data/polyploid.chr22.phased.vcf", phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "chr22"
    assert table.samples == ["HG00514_NA19240"]
    assert len(table.variants) == 8

    expected_phase = [
        VariantCallPhase(block_id=20000000, phase=(1, 0, 1, 1), quality=None),
        VariantCallPhase(block_id=20000000, phase=(1, 0, 1, 0), quality=None),
        None,
        VariantCallPhase(block_id=20000000, phase=(1, 0, 1, 1), quality=None),
        VariantCallPhase(block_id=20001000, phase=(1, 0, 1, 1), quality=None),
        VariantCallPhase(block_id=20001000, phase=(0, 0, 0, 1), quality=None),
        VariantCallPhase(block_id=20001000, phase=(0, 0, 0, 1), quality=None),
        VariantCallPhase(block_id=20001000, phase=(0, 0, 0, 1), quality=None),
    ]
    print("Got:")
    for variant in table.phases[0]:
        print(variant)
    print("Exp:")
    for variant in expected_phase:
        print(variant)
    assert list(table.phases[0]) == expected_phase


def test_read_tetraploid_genotype_likelihoods():
    tables = list(
        VcfReader(
            "tests/data/polyploid.chr22.unphased.vcf", phases=False, genotype_likelihoods=True
        )
    )
    assert len(tables) == 1
    table = tables[0]
    assert len(table.variants) == 8
    exp_gl = [
        GenotypeLikelihoods([-x / 10 for x in [19, 28, 29, 2, 10, 6]]),
        GenotypeLikelihoods([-x / 10 for x in [1, 8, 29, 24, 15, 23]]),
        GenotypeLikelihoods([-x / 10 for x in [25, 33, 35, 31, 0, 30]]),
        GenotypeLikelihoods([-x / 10 for x in [6, 27, 6, 3, 46, 42]]),
    ] * 2
    assert table.genotype_likelihoods_of(table.samples[0]) == exp_gl


def test_unsupported_ploidy():
    try:
        _ = list(VcfReader("tests/data/hexadecaploid.chr22.vcf", phases=False))
    except PloidyError:
        return
    assert False


def test_unsupported_ploidy_phased():
    try:
        _ = list(VcfReader("tests/data/hexadecaploid.chr22.vcf", phases=True))
    except PloidyError:
        return
    assert False


def test_inconsistent_ploidy():
    try:
        _ = list(VcfReader("tests/data/polyploid.chr22.inconsistent.vcf", phases=False))
    except PloidyError:
        return
    assert False


def test_inconsistent_ploidy_phased():
    try:
        _ = list(VcfReader("tests/data/polyploid.chr22.inconsistent.vcf", phases=True))
    except PloidyError:
        return
    assert False


def test_vcf_without_index(tmp_path):
    vcf_path = tmp_path / "file.vcf.gz"
    import shutil

    shutil.copy("tests/data/haplotag_1.vcf.gz", vcf_path)
    with raises(VcfIndexMissing):
        with VcfReader(vcf_path) as vr:
            list(vr.fetch("chr1"))



"""
Pedigree-related functions
"""
from abc import ABC, abstractmethod

import math
from pathlib import Path

from typing import Optional, Union, Sequence, List, IO, Iterator, Mapping
from collections import Counter, defaultdict
from dataclasses import dataclass
import logging

from whatshap.core import Genotype

logger = logging.getLogger(__name__)


class ParseError(Exception):
    pass


@dataclass
class RecombinationMapEntry:
    position: int
    cum_distance: float


@dataclass(order=True)
class RecombinationEvent:
    position1: int
    position2: int
    transmitted_hap_father1: int
    transmitted_hap_father2: int
    transmitted_hap_mother1: int
    transmitted_hap_mother2: int
    recombination_cost: float


def _interpolate(
    point: int, start_pos: int, end_pos: int, start_value: float, end_value: float
) -> float:

    assert start_pos <= point <= end_pos
    if start_pos == point == end_pos:
        assert start_value == end_value
        return start_value
    return start_value + ((point - start_pos) * (end_value - start_value) / (end_pos - start_pos))


MINIMUM_GENETIC_DISTANCE: float = 1e-10  # cM


def recombination_cost_map(
    genetic_map: Sequence[RecombinationMapEntry], positions: Sequence[int]
) -> Sequence[int]:

    assert len(genetic_map) > 0

    # Step 1: compute cumulative genetic distances from start of chromosome
    #         to each position.
    cumulative_distances: List[float] = []
    # i and j are such that genetic_map[i].position <= position <= genetic_map[j].position
    # i and j are None if no such values exist (because we are at the end of the list)
    i: Optional[int] = None
    j: Optional[int] = 0

    for position in positions:
        # update i to meet the invariant
        if (i is None) and (genetic_map[0].position <= position):
            i = 0
        while (
            (i is not None)
            and (i + 1 < len(genetic_map))
            and (genetic_map[i + 1].position <= position)
        ):
            i += 1

        # update j to meet the invariant
        while (j is not None) and (genetic_map[j].position < position):
            if j + 1 < len(genetic_map):
                j += 1
            else:
                j = None

        # interpolate
        if i is None:
            assert j is not None
            d = _interpolate(position, 0, genetic_map[j].position, 0, genetic_map[j].cum_distance)
        elif j is None:
            # Point outside the genetic map --> extrapolating using average recombination rate
            avg_rate = genetic_map[-1].cum_distance / genetic_map[-1].position
            d = genetic_map[-1].cum_distance + (position - genetic_map[-1].position) * avg_rate
        else:
            assert genetic_map[i].position <= position <= genetic_map[j].position
            d = _interpolate(
                position,
                genetic_map[i].position,
                genetic_map[j].position,
                genetic_map[i].cum_distance,
                genetic_map[j].cum_distance,
            )
        cumulative_distances.append(d)

    # Step 2: compute costs (= phred-scaled recombination probabilities between two positions)
    result = [0]
    for i in range(1, len(cumulative_distances)):
        d = cumulative_distances[i] - cumulative_distances[i - 1]
        d = max(d, MINIMUM_GENETIC_DISTANCE)
        result.append(round(centimorgen_to_phred(d)))

    return result


def centimorgen_to_phred(distance: float) -> float:
    assert distance >= 0
    if distance == 0:
        raise ValueError("Cannot convert genetic distance of zero to phred.")
    elif distance < 1e-10:
        return -10.0 * (math.log10(distance) - 2.0)
    else:
        p = (1.0 - math.exp(-(2.0 * distance) / 100.0)) / 2.0
        return -10.0 * math.log10(p)


def mendelian_conflict(genotypem: Genotype, genotypef: Genotype, genotypec: Genotype) -> bool:
    # TODO: Maybe inefficient
    alleles_m = genotypem.as_vector()
    alleles_f = genotypef.as_vector()
    alleles_c = genotypec.as_vector()
    if alleles_c[0] in alleles_m and alleles_c[1] in alleles_f:
        return False
    elif alleles_c[1] in alleles_m and alleles_c[0] in alleles_f:
        return False
    else:
        return True


def find_recombination(
    transmission_vector: Sequence[int],
    components: Mapping[int, int],
    positions: Sequence[int],
    recombcost: Sequence[int],
) -> Sequence[RecombinationEvent]:

    assert len(transmission_vector) == len(positions) == len(recombcost)
    assert set(components.keys()).issubset(set(positions))
    position_to_index = {pos: i for i, pos in enumerate(positions)}
    blocks = defaultdict(list)
    for position, block_id in components.items():
        blocks[block_id].append(position)

    events = []
    cum_recomb_cost = 0
    for block_id, block in blocks.items():
        block.sort()
        block_transmission_vector = [transmission_vector[position_to_index[i]] for i in block]
        block_recomb_cost = [recombcost[position_to_index[i]] for i in block]
        if len(block) <= 2:
            continue
        for i in range(2, len(block)):
            if block_transmission_vector[i - 1] != block_transmission_vector[i]:
                events.append(
                    RecombinationEvent(
                        block[i - 1],
                        block[i],
                        block_transmission_vector[i - 1] % 2,
                        block_transmission_vector[i] % 2,
                        block_transmission_vector[i - 1] // 2,
                        block_transmission_vector[i] // 2,
                        block_recomb_cost[i],
                    )
                )
                cum_recomb_cost += block_recomb_cost[i]

    logger.info("Cost accounted for by recombination events: %d", cum_recomb_cost)
    events.sort()
    return events


class RecombinationCostComputer(ABC):
    @abstractmethod
    def compute(self, positions: Sequence[int]) -> Sequence[int]:
        pass


class GeneticMapRecombinationCostComputer(RecombinationCostComputer):
    def __init__(self, genetic_map_path):
        self._genetic_map = self.load_genetic_map(genetic_map_path)

    @staticmethod
    def load_genetic_map(filename: Union[str, Path]) -> Sequence[RecombinationMapEntry]:
        # Example:
        #
        # position COMBINED_rate(cM/Mb) Genetic_Map(cM)
        # 55550 0 0
        # 721290 2.685807669 0.410292036939447
        #
        genetic_map = []
        warned_zero_distance = False
        with open(filename) as fid:
            for line_number, line in enumerate(fid, 1):
                if line_number == 1:
                    continue
                fields = line.strip().split()
                if not fields:
                    # Skip empty lines
                    continue
                if len(fields) != 3:
                    raise ParseError(
                        f"Error at line {line_number} of genetic map file '{filename}': "
                        f"Found {len(fields)} fields instead of 3"
                    )
                try:
                    position = int(fields[0])
                    cum_distance = float(fields[2])
                except ValueError as e:
                    raise ParseError(
                        f"Error at line {line_number} of genetic map file '{filename}': {e}"
                    )
                genetic_map.append(
                    RecombinationMapEntry(position=position, cum_distance=cum_distance)
                )
                if len(genetic_map) >= 2:
                    if not warned_zero_distance and (
                        genetic_map[-2].cum_distance == genetic_map[-1].cum_distance
                    ):
                        logger.warning("Zero genetic distances encountered in %s", filename)
                        warned_zero_distance = True

        return genetic_map

    def compute(self, positions: Sequence[int]) -> Sequence[int]:
        return recombination_cost_map(self._genetic_map, positions)


class UniformRecombinationCostComputer(RecombinationCostComputer):
    def __init__(self, recombination_rate: float):
        self._recombination_rate = recombination_rate

    @staticmethod
    def uniform_recombination_map(recombrate: float, positions) -> Sequence[int]:
        """
        For a list of positions and a constant recombination rate (in cM/Mb),
        return a list "results" of the same length as "positions" such that
        results[i] is the phred-scaled recombination probability between
        positions[i-1] and positions[i].
        """
        return [0] + [
            round(centimorgen_to_phred((positions[i] - positions[i - 1]) * 1e-6 * recombrate))
            for i in range(1, len(positions))
        ]

    def compute(self, positions: Sequence[int]) -> Sequence[int]:
        return self.uniform_recombination_map(self._recombination_rate, positions)


@dataclass
class Trio:
    """Relationships are modelled as a set of trios (mother, father, child)."""

    child: str
    father: Optional[str]
    mother: Optional[str]


class PedReader:
    """
    A parser for PED/FAM files as used by PLINK and other tools.

    According to <http://pngu.mgh.harvard.edu/~purcell/plink/data.shtml>:
    The PED file is a white-space (space or tab) delimited file: the first six
    columns are mandatory:
    * Family ID
    * Individual ID
    * Paternal ID
    * Maternal ID
    * Sex (1=male; 2=female; other=unknown)
    * Phenotype

    All fields except the individual, maternal and paternal ID are ignored by
    this class. The entire file is read upon construction.
    """

    def __init__(self, file: Union[str, Path, IO]):
        if isinstance(file, (str, Path)):
            with open(file) as f:
                self.trios = self._parse(f)
        else:
            self.trios = self._parse(file)

    @staticmethod
    def _parse_record(line: str) -> Trio:
        """
        Parse a single non-comment line of a PED or FAM file.
        """
        fields = line.split()
        if len(fields) < 6:
            raise ParseError("Less than six fields found in PED/FAM file")
        individual_id, paternal_id, maternal_id = fields[1:4]
        return Trio(
            child=individual_id,
            father=paternal_id if paternal_id != "0" else None,
            mother=maternal_id if maternal_id != "0" else None,
        )

    def _parse(self, file: IO) -> Sequence[Trio]:
        trios = []
        for line in file:
            if line.startswith("#") or line == "\n":
                continue
            trios.append(self._parse_record(line))
        self._sanity_check(trios)
        return trios

    @staticmethod
    def _sanity_check(trios: Sequence[Trio]) -> None:
        """
        Ensure that each individual occurs only once in the file.
        """
        children = [trio.child for trio in trios]
        if not children:
            return
        id, count = Counter(children).most_common()[0]
        if count > 1:
            raise ParseError(f"Individual {id!r} occurs more than once in PED file")

    def __iter__(self) -> Iterator[Trio]:
        return iter(self.trios)

    def samples(self) -> Sequence[str]:
        """Return a list of all mentioned individuals"""
        samples = set()
        for trio in self.trios:
            if trio.child is None or trio.mother is None or trio.father is None:
                continue
            samples.add(trio.father)
            samples.add(trio.mother)
            samples.add(trio.child)
        return list(samples)



# Sphinx configuration

import os
import sys
import time

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath(os.pardir))

extensions = [
    "sphinx.ext.mathjax",
    "sphinx_issues",
    "sphinx_better_subsection",
]

templates_path = ["_templates"]

source_suffix = ".rst"

master_doc = "index"

authors = "Murray Patterson, Alexander Schönhuth, Tobias Marschall, Marcel Martin and WhatsHap contributors"
project = "whatshap"
copyright = "2014-{}, {}".format(time.gmtime().tm_year, authors)

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
from pkg_resources import get_distribution

release = get_distribution("whatshap").version
# Read The Docs modifies the conf.py script and we therefore get
# version numbers like 0.12+0.g27d0d31
if os.environ.get("READTHEDOCS") == "True":
    version = ".".join(release.split(".")[:2])
else:
    version = release

release = version

issues_github_path = "whatshap/whatshap"

exclude_patterns = ["_build"]

# The reST default role (used for this markup: `text`) to use for all
# documents.
# default_role = None

pygments_style = "sphinx"

html_theme = "default"

html_static_path = ["_static"]
